# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import time
import warnings
from collections import namedtuple
from uuid import UUID

from data_pipeline._avro_payload import _AvroPayload
from data_pipeline._encryption_helper import EncryptionHelper
from data_pipeline._fast_uuid import FastUUID
from data_pipeline.config import get_config
from data_pipeline.envelope import Envelope
from data_pipeline.helpers.lists import unlist
from data_pipeline.helpers.yelp_avro_store import _AvroStringStore
from data_pipeline.message_type import _ProtectedMessageType
from data_pipeline.message_type import MessageType
from data_pipeline.meta_attribute import MetaAttribute
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer


logger = get_config().logger


KafkaPositionInfo = namedtuple('KafkaPositionInfo', [
    'offset',               # Offset of the message in the topic
    'partition',            # Partition of the topic the message was from
    'key'                   # Key of the message, may be `None`
])


PayloadFieldDiff = namedtuple('PayloadFieldDiff', [
    'old_value',            # Value of the field before update
    'current_value'         # Value of the field after update
])


class MissingMetaAttributeException(Exception):
    def __init__(self, schema_id, meta_ids, mandatory_meta_ids):
        Exception.__init__(
            self,
            "Meta Attributes with IDs `{0}` are not found for schema_id "
            "`{1}`.".format(
                ", ".join(str(m) for m in (mandatory_meta_ids - meta_ids)),
                schema_id
            ))


class NoEntryPayload(object):
    """ This class denotes that no previous value exists for the field. """
    pass


class InvalidOperation(Exception):
    pass


class Message(object):
    """Encapsulates a data pipeline message with metadata about the message.

    Validates metadata, but not the payload itself. This class is not meant
    to be used directly. Use specific type message class instead:
    :class:`data_pipeline.message.CreateMessage`,
    :class:`data_pipeline.message.UpdateMessage`,
    :class:`data_pipeline.message.DeleteMessage`, and
    :class:`data_pipeline.message.RefreshMessage`.

    Args:
        schema_id (int): Identifies the schema used to encode the payload.
        reader_schema_id (Optional[int]): Identifies the schema used to decode
            the payload.
        topic (Optional[str]): Kafka topic to publish into.  It is highly
            recommended to leave it unassigned and let the Schematizer decide
            the topic of the schema.  Use caution when overriding the topic.
        payload (bytes): Avro-encoded message - encoded with schema identified
            by `schema_id`. This is expected to be None for messages on their
            way to being published. Either `payload` or `payload_data` must be
            provided but not both.
        payload_data: The contents of message, which will be lazily
            encoded with schema identified by `schema_id`. Either `payload` or
            `payload_data` must be provided but not both. Type of payload_data
            should match the avro type specified schema.
        uuid (bytes, optional): Globally-unique 16-byte identifier for the
            message.  A uuid4 will be generated automatically if this isn't
            provided.
        contains_pii (bool, optional): Indicates that the payload contains PII,
            so the clientlib can properly encrypt the data and mark it as
            sensitive, defaults to False. The data pipeline consumer will
            automatically decrypt fields containing PII. This field should not
            be used to indicate that a topic should be encrypted, because
            PII information will be used to indicate to various systems how
            to handle the data, in addition to automatic decryption.
        timestamp (int, optional): A unix timestamp for the message.  If this is
            not provided, a timestamp will be generated automatically.  If the
            message is coming directly from an upstream source, and the
            modification time is available in that source, it's appropriate to
            use that timestamp.  Otherwise, it's probably best to have the
            timestamp represent when the message was generated.  If the message
            is derived from an upstream data pipeline message, reuse the
            timestamp from that upstream message.

            Timestamp is used internally by the clientlib to monitor timings and
            other metadata about the data pipeline as a system.
            Consequently, there is no need to store information about when this
            message passed through individual systems in the message itself,
            as it is otherwise recorded.  See DATAPIPE-169 for details about
            monitoring.
        upstream_position_info (dict, optional): This dict must only contain
            primitive types.  It is not used internally by the data pipeline,
            so the content is left to the application.  The clientlib will
            track these objects and provide them back from the producer to
            identify the last message that was successfully published, both
            overall and per topic.
        keys (tuple, optional): This should either be a tuple of strings
            or None.  If it's a tuple of strings, the clientlib will combine
            those strings and use them as key when publishing into Kafka.
        dry_run (boolean): When set to True, Message will return a string
            representation of the payload and previous payload, instead of
            the avro encoded message.  This is to avoid loading the schema
            from the schema store.  Defaults to False.
        meta (list of MetaAttribute, optional): This should be a list of
            MetaAttribute objects or None. This is used to contain information
            about metadata. These meta attributes are serialized using their
            respective avro schema, which is registered with the schematizer.
            Hence meta should be set with a dict which contains schema_id and
            payload as keys to construct the MetaAttribute objects. The
            payload is deserialized using the schema_id.

    Remarks:
        Although `previous_payload` and `previous_payload_data` are not
        applicable and do not exist in non-update type Message classes,
        these classes do not prevent them from being added dynamically.
        Ensure not to use these attributes for non-update type Message classes.
    """

    _message_type = None
    """Identifies the nature of the message. The valid value is one of the
    data_pipeline.message_type.MessageType. It must be set by child class.
    """

    _fast_uuid = FastUUID()
    """UUID generator - this isn't a @cached_property so it can be serialized"""

    @property
    def _schematizer(self):
        return get_schematizer()

    @property
    def topic(self):
        return self._topic

    def _set_topic(self, topic):
        if not isinstance(topic, str):
            raise TypeError("Topic must be a non-empty string")
        if len(topic) == 0:
            raise ValueError("Topic must be a non-empty string")
        self._topic = topic

    @property
    def schema_id(self):
        return self._avro_payload.schema_id

    @property
    def reader_schema_id(self):
        return self._avro_payload.reader_schema_id

    @property
    def message_type(self):
        """Identifies the nature of the message."""
        return self._message_type

    @property
    def uuid(self):
        return self._uuid

    def _set_uuid(self, uuid):
        if uuid is None:
            # UUID generation is expensive.  Using FastUUID instead of the built
            # in UUID methods increases Messages that can be instantiated per
            # second from ~25,000 to ~185,000.  Not generating UUIDs at all
            # increases the throughput further still to about 730,000 per
            # second.
            uuid = self._fast_uuid.uuid4()
        elif len(uuid) != 16:
            raise TypeError(
                "UUIDs should be exactly 16 bytes.  Conforming UUID's can be "
                "generated with `import uuid; uuid.uuid4().bytes`."
            )
        self._uuid = uuid

    @property
    def uuid_hex(self):
        # TODO: DATAPIPE-848
        return UUID(bytes=self.uuid).hex

    @property
    def contains_pii(self):
        if self._contains_pii is not None:
            return self._contains_pii
        self._set_contains_pii()
        return self._contains_pii

    def _set_contains_pii(self):
        self._contains_pii = self._schematizer.get_schema_by_id(
            self.schema_id
        ).topic.contains_pii

    @property
    def encryption_type(self):
        self._set_encryption_type_if_necessary()
        return self._encryption_type

    def _set_encryption_type_if_necessary(self):
        if self._encryption_type or not self._should_be_encrypted:
            return
        config_encryption_type = get_config().encryption_type
        if config_encryption_type is None:
            raise ValueError(
                "Encryption type must be set when message requires to be encrypted."
            )
        self._encryption_type = config_encryption_type
        self._encryption_helper = EncryptionHelper(config_encryption_type)
        self._set_encryption_meta()

    @property
    def _should_be_encrypted(self):
        """Whether this message should be encrypted.  So far the criteria used
        to determine if the message should be encrypted is the pii information.
        Include additional criteria if necessary.
        """
        if self._should_be_encrypted_state is not None:
            return self._should_be_encrypted_state

        self._should_be_encrypted_state = self.contains_pii
        return self._should_be_encrypted_state

    def _set_encryption_meta(self):
        if self._meta is None:
            self._meta = []
        self._pop_encryption_meta(self._encryption_type, self._meta)
        self._meta.append(self._encryption_helper.encryption_meta)

    @property
    def dry_run(self):
        return self._avro_payload.dry_run

    @property
    def meta(self):
        self._set_encryption_type_if_necessary()
        return self._meta

    def _set_meta(self, meta, schema_id):
        if (not self._is_valid_optional_type(meta, list) or
                self._any_invalid_type(meta, MetaAttribute)):
            raise TypeError("Meta must be None or list of MetaAttribute objects.")
        meta_attr_schema_ids = {
            meta_attr.schema_id for meta_attr in meta
        } if meta else set()
        mandatory_meta_ids = set(
            self._schematizer.get_meta_attributes_by_schema_id(schema_id)
        )
        if not mandatory_meta_ids.issubset(meta_attr_schema_ids):
            raise MissingMetaAttributeException(
                schema_id,
                meta_attr_schema_ids,
                mandatory_meta_ids
            )
        self._meta = meta

    def get_meta_attr_by_type(self, meta, meta_type):
        if meta is not None:
            attributes_with_type = [m for m in meta if m.source == meta_type]
            return unlist(attributes_with_type)
        return None

    def _get_meta_attr_avro_repr(self):
        if self.meta is not None:
            return [meta_attr.avro_repr for meta_attr in self.meta]
        return None

    @property
    def timestamp(self):
        return self._timestamp

    def _set_timestamp(self, timestamp):
        if timestamp is None:
            timestamp = int(time.time())
        self._timestamp = timestamp

    @property
    def upstream_position_info(self):
        return self._upstream_position_info

    def _set_upstream_position_info(self, upstream_position_info):
        if not self._is_valid_optional_type(upstream_position_info, dict):
            raise TypeError("upstream_position_info must be None or a dict")
        self._upstream_position_info = upstream_position_info

    @upstream_position_info.setter
    def upstream_position_info(self, upstream_position_info):
        # This should be the only exception that users can update the data after
        # the message is created. The `upstream_position_info` is not used in
        # the data pipeline and the data is up to the application, so it should
        # be ok. It is more efficient and simpler to allow application updating
        # the data than creating new instance with new data each time.
        self._set_upstream_position_info(upstream_position_info)

    @property
    def kafka_position_info(self):
        """The kafka offset, partition, and key of the message if it
        was consumed from kafka. This is expected to be None for messages
        on their way to being published.
        """
        return self._kafka_position_info

    def _set_kafka_position_info(self, kafka_position_info):
        if not self._is_valid_optional_type(kafka_position_info, KafkaPositionInfo):
            raise TypeError(
                "kafka_position_info must be None or a KafkaPositionInfo"
            )
        self._kafka_position_info = kafka_position_info

    @property
    def keys(self):
        """Currently this support primary keys for flat record
        type avro schema. Support for primary keys in nested
        avro schema will be handled in future versions.
        """
        if self._keys is not None:
            return self._keys
        self._set_keys()
        return self._keys

    def _set_keys(self):
        avro_schema = self._schematizer.get_schema_by_id(self.schema_id)
        self._keys = {
            key: self.payload_data[key] for key in avro_schema.primary_keys
        }

    @property
    def encoded_keys(self):
        writer = _AvroStringStore().get_writer(
            id_key="{0}_{1}".format("keys", self.schema_id),
            avro_schema=self._keys_avro_json
        )
        return writer.encode(message_avro_representation=self.keys)

    def _extract_key_fields(self):
        avro_schema = self._schematizer.get_schema_by_id(
            self.schema_id
        )
        schema_json = avro_schema.schema_json

        fields = schema_json.get('fields', [])
        field_name_to_field = {f['name']: f for f in fields}
        key_fields = [field_name_to_field[pkey] for pkey in avro_schema.primary_keys]
        return key_fields

    @property
    def _keys_avro_json(self):
        return {
            "type": "record",
            "namespace": "yelp.data_pipeline",
            "name": "primary_keys",
            "doc": "Represents primary keys present in Message payload.",
            "fields": self._extract_key_fields()
        }

    @property
    def payload(self):
        return self._avro_payload.payload

    @property
    def payload_data(self):
        return self._avro_payload.payload_data

    @property
    def payload_diff(self):
        return {
            field: self._get_field_diff(field) for field in self.payload_data
        }

    def __init__(
        self,
        schema_id,
        reader_schema_id=None,
        topic=None,
        payload=None,
        payload_data=None,
        uuid=None,
        contains_pii=None,
        timestamp=None,
        upstream_position_info=None,
        kafka_position_info=None,
        keys=None,
        dry_run=False,
        meta=None
    ):
        # The decision not to just pack the message, but to validate it, is
        # intentional here.  We want to perform more sanity checks than avro
        # does, and in addition, this check is quite a bit faster than
        # serialization.  Finally, if we do it this way, we can lazily
        # serialize the payload in a subclass if necessary.

        # TODO(DATAPIPE-416|psuben): Make it so contains_pii is no longer
        # overrideable. Now the pass-in contains_pii is no longer used. Next
        # is to remove it from the function signature altogether.
        if contains_pii is not None:
            warnings.simplefilter('always', DeprecationWarning)
            warnings.warn(
                "contains_pii is deprecated. Please stop passing it in.",
                DeprecationWarning
            )
        if topic:
            warnings.simplefilter("always", category=DeprecationWarning)
            warnings.warn("Passing in topics explicitly is deprecated.", DeprecationWarning)
        self._avro_payload = _AvroPayload(
            schema_id=schema_id,
            reader_schema_id=reader_schema_id,
            payload=payload,
            payload_data=payload_data,
            dry_run=dry_run
        )
        self._set_topic(
            topic or str(self._schematizer.get_schema_by_id(schema_id).topic.name)
        )
        self._set_uuid(uuid)
        self._set_timestamp(timestamp)
        self._set_upstream_position_info(upstream_position_info)
        self._set_kafka_position_info(kafka_position_info)
        if keys is not None:
            warnings.simplefilter("always", category=DeprecationWarning)
            warnings.warn("Passing in keys explicitly is deprecated.", DeprecationWarning)
        self._keys = None
        self._set_meta(meta, schema_id)
        self._should_be_encrypted_state = None
        self._encryption_type = None
        self._contains_pii = None

    def _is_valid_optional_type(self, value, typ):
        return value is None or isinstance(value, typ)

    def _any_invalid_type(self, value_list, typ):
        if not value_list:
            return False
        return any(not isinstance(value, typ) for value in value_list)

    def _encrypt_payload_if_necessary(self, payload):
        if self.encryption_type is not None:
            return self._encryption_helper.encrypt_payload(payload)
        return payload

    @property
    def avro_repr(self):
        return {
            'uuid': self.uuid,
            'message_type': self.message_type.name,
            'schema_id': self.schema_id,
            'payload': self._encrypt_payload_if_necessary(self.payload),
            'timestamp': self.timestamp,
            'meta': self._get_meta_attr_avro_repr(),
            'encryption_type': self.encryption_type,
        }

    @classmethod
    def create_from_unpacked_message(
        cls,
        unpacked_message,
        reader_schema_id=None,
        kafka_position_info=None
    ):
        encryption_type = unpacked_message['encryption_type']
        meta = cls._get_unpacked_meta(unpacked_message)
        encryption_meta = cls._pop_encryption_meta(encryption_type, meta)
        payloads = {
            param_name: cls._get_unpacked_decrypted_payload(
                payload,
                encryption_type=encryption_type,
                encryption_meta=encryption_meta
            )
            for param_name, payload in cls._get_all_payloads(
                unpacked_message
            ).iteritems()
        }

        message_params = {
            'uuid': unpacked_message['uuid'],
            'schema_id': unpacked_message['schema_id'],
            'reader_schema_id': reader_schema_id,
            'timestamp': unpacked_message['timestamp'],
            'meta': meta,
            'kafka_position_info': kafka_position_info
        }
        message_params.update(payloads)
        message = cls(**message_params)
        message._should_be_encrypted_state = bool(encryption_type)
        return message

    @classmethod
    def _get_unpacked_meta(cls, unpacked_message):
        return [
            MetaAttribute(schema_id=o['schema_id'], payload=o['payload'])
            for o in unpacked_message['meta']
        ] if unpacked_message['meta'] else None

    @classmethod
    def _get_unpacked_decrypted_payload(
        cls,
        payload,
        encryption_type,
        encryption_meta
    ):
        if not encryption_type:
            return payload

        encryption_helper = EncryptionHelper(encryption_type, encryption_meta)
        return encryption_helper.decrypt_payload(payload)

    @classmethod
    def _pop_encryption_meta(cls, encryption_type, meta):
        if not encryption_type or not meta:
            return None

        encryption_meta = EncryptionHelper.get_encryption_meta_by_encryption_type(
            encryption_type
        )
        for index, meta_attr in enumerate(meta):
            if meta_attr.schema_id == encryption_meta.schema_id:
                target_meta = meta_attr
                meta[index] = meta[-1]
                meta.pop()
                return target_meta
        return None

    @classmethod
    def _get_all_payloads(cls, unpacked_message):
        """Get all the payloads in the message."""
        return {'payload': unpacked_message['payload']}

    def _get_cleaned_pii_data(self, data):
        if not isinstance(data, dict):
            return unicode(type(data))
        return {
            key: self._get_cleaned_pii_data(value)
            for key, value in data.iteritems()
        }

    def reload_data(self):
        """Populate the payload data or the payload if it hasn't done so.
        """
        self._avro_payload.reload_data()

    @property
    def _str_repr(self):
        cleaned_payload_data = self.payload_data
        if self.contains_pii:
            cleaned_payload_data = self._get_cleaned_pii_data(
                self._avro_payload.printable_payload_data
            )
        return {
            'uuid': self.uuid_hex,
            'message_type': self.message_type.name,
            'schema_id': self.schema_id,
            'payload_data': cleaned_payload_data,
            'timestamp': self.timestamp,
            'meta': None if self.meta is None else [m._asdict() for m in self.meta],
            'encryption_type': self.encryption_type
        }

    def __str__(self):
        return str(self._str_repr)

    def __eq__(self, other):
        return type(self) is type(other) and self._eq_key == other._eq_key

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._eq_key)

    @property
    def _eq_key(self):
        """Returns a tuple representing a unique key for this Message.

        Note:
            We don't include `payload_data` in the key tuple as we should be
            confident that if `payload` matches then `payload_data` will as
            well, and there is an extra overhead from decoding.
        """
        return (
            self.message_type,
            self.topic,
            self.schema_id,
            self.payload,
            self.uuid,
            self.timestamp,
            self.upstream_position_info,
            self.kafka_position_info,
            self.dry_run,
            self.encryption_type
        )


class CreateMessage(Message):

    _message_type = MessageType.create

    def _get_field_diff(self, field):
        return PayloadFieldDiff(
            old_value=NoEntryPayload,
            current_value=self.payload_data[field]
        )


class DeleteMessage(Message):

    _message_type = MessageType.delete

    def _get_field_diff(self, field):
        return PayloadFieldDiff(
            old_value=self.payload_data[field],
            current_value=NoEntryPayload
        )


class RefreshMessage(Message):

    _message_type = MessageType.refresh

    def _get_field_diff(self, field):
        raise InvalidOperation()


class LogMessage(Message):
    _message_type = MessageType.log

    def _get_field_diff(self, field):
        raise InvalidOperation()


class MonitorMessage(Message):
    _message_type = _ProtectedMessageType.monitor

    def _get_field_diff(self, field):
        raise InvalidOperation()


class RegistrationMessage(Message):
    _message_type = _ProtectedMessageType.registration

    def _get_field_diff(self, field):
        raise InvalidOperation()


class UpdateMessage(Message):
    """Message for update type. This type of message requires previous
    payload in addition to the payload.

    For complete argument docs, see :class:`data_pipeline.message.Message`.

    Args:
        previous_payload (bytes): Avro-encoded message - encoded with schema
            identified by `schema_id`  Required when message type is
            MessageType.update.  Either `previous_payload` or
            `previous_payload_data` must be provided but not both.
        previous_payload_data (dict): The contents of message, which will be
            lazily encoded with schema identified by `schema_id`.  Required
            when message type is MessageType.update.  Either `previous_payload`
            or `previous_payload_data` must be provided but not both.
    """

    _message_type = MessageType.update

    def __init__(
        self,
        schema_id,
        reader_schema_id=None,
        topic=None,
        payload=None,
        payload_data=None,
        previous_payload=None,
        previous_payload_data=None,
        uuid=None,
        contains_pii=None,
        timestamp=None,
        upstream_position_info=None,
        kafka_position_info=None,
        keys=None,
        dry_run=False,
        meta=None,
    ):
        super(UpdateMessage, self).__init__(
            schema_id,
            reader_schema_id,
            topic=topic,
            payload=payload,
            payload_data=payload_data,
            uuid=uuid,
            contains_pii=contains_pii,
            timestamp=timestamp,
            upstream_position_info=upstream_position_info,
            kafka_position_info=kafka_position_info,
            keys=keys,
            dry_run=dry_run,
            meta=meta,
        )
        self._previous_avro_payload = _AvroPayload(
            schema_id=schema_id,
            reader_schema_id=reader_schema_id,
            payload=previous_payload,
            payload_data=previous_payload_data,
            dry_run=dry_run
        )

    @property
    def _eq_key(self):
        """Returns a tuple representing a unique key for this Message.

        Note:
            We don't include `previous_payload_data` in the key as we should
            be confident that if `previous_payload` matches then
            `previous_payload_data` will as well, and there is an extra
            overhead from decoding.
        """
        return super(UpdateMessage, self)._eq_key + (self.previous_payload,)

    @property
    def previous_payload(self):
        """Avro-encoded message - encoded with schema identified by
        `schema_id`.  Required when message type is `MessageType.update`.
        """
        return self._previous_avro_payload.payload

    @property
    def previous_payload_data(self):
        return self._previous_avro_payload.payload_data

    @property
    def avro_repr(self):
        repr_dict = super(UpdateMessage, self).avro_repr
        repr_dict['previous_payload'] = self._encrypt_payload_if_necessary(
            self.previous_payload
        )
        return repr_dict

    @classmethod
    def _get_all_payloads(cls, unpacked_message):
        """Get all the payloads in the message."""
        return {
            'payload': unpacked_message['payload'],
            'previous_payload': unpacked_message['previous_payload']
        }

    def reload_data(self):
        """Populate the previous payload data or decode the previous payload
        if it hasn't done so. The payload encoding/payload data decoding is
        taken care of by the `Message.reload` function in the parent class.
        """
        super(UpdateMessage, self).reload_data()
        self._previous_avro_payload.reload_data()

    def _has_field_changed(self, field):
        return self.payload_data[field] != self.previous_payload_data[field]

    def _get_field_diff(self, field):
        return PayloadFieldDiff(
            old_value=self.previous_payload_data[field],
            current_value=self.payload_data[field]
        )

    @property
    def has_changed(self):
        return any(self._has_field_changed(field) for field in self.payload_data)

    @property
    def payload_diff(self):
        return {
            field: self._get_field_diff(field)
            for field in self.payload_data if self._has_field_changed(field)
        }

    @property
    def _str_repr(self):
        # Calls the _str_repr from the super class resulting in encrypted pii data.
        repr_dict = super(UpdateMessage, self)._str_repr
        repr_dict['previous_payload_data'] = (
            self.previous_payload_data if not self.contains_pii
            else self._get_cleaned_pii_data(self.previous_payload_data)
        )

        return repr_dict


_message_type_to_class_map = {
    o._message_type.name: o for o in Message.__subclasses__() if o._message_type
}


def create_from_kafka_message(
    kafka_message,
    envelope=None,
    force_payload_decoding=True,
    reader_schema_id=None
):
    """ Build a data_pipeline.message.Message from a yelp_kafka message. If no
    reader schema id is provided, the schema used for encoding will be used for
    decoding.

    Args:
        kafka_message (kafka.common.KafkaMessage): The message info which
            has the topic, partition, offset, key, and value(payload) of
            the received message.
        envelope (Optional[:class:data_pipeline.envelope.Envelope]): Envelope
            instance that unpacks the data pipeline messages.
        force_payload_decoding (Optional[boolean]): If this is set to `True`
            then we will decode the payload/previous_payload immediately.
            Otherwise the decoding will happen whenever the lazy *_data
            properties are accessed.
        reader_schema_id (Optional[int]): Schema id used to decode the
            kafka_message and build data_pipeline.message.Message message.
            Defaults to None.


    Returns (class:`data_pipeline.message.Message`):
        The message object
    """
    kafka_position_info = KafkaPositionInfo(
        offset=kafka_message.offset,
        partition=kafka_message.partition,
        key=kafka_message.key,
    )
    return _create_message_from_packed_message(
        packed_message=kafka_message,
        envelope=envelope or Envelope(),
        force_payload_decoding=force_payload_decoding,
        kafka_position_info=kafka_position_info,
        reader_schema_id=reader_schema_id
    )


def create_from_offset_and_message(
    offset_and_message,
    force_payload_decoding=True,
    reader_schema_id=None,
    envelope=None
):
    """
    Build a data_pipeline.message.Message from a kafka.common.OffsetAndMessage.
    If no reader schema id is provided, the schema used for encoding will be
    used for decoding.

    Args:
        offset_and_message (kafka.common.OffsetAndMessage): a namedtuple
            containing the offset and message. Message contains magic,
            attributes, keys and values.
        force_payload_decoding (Optional[boolean]): If this is set to `True`
            then we will decode the payload/previous_payload immediately.
            Otherwise the decoding will happen whenever the lazy *_data
            properties are accessed.
        reader_schema_id (Optional[int]): Schema id used to decode the incoming
            kafka message and build data_pipeline.message.Message message.
            Defaults to None.
        envelope (Optional[:class:data_pipeline.envelope.Envelope]): Envelope
            instance that unpacks the data pipeline messages.

    Returns (data_pipeline.message.Message):
        The message object
    """
    return _create_message_from_packed_message(
        packed_message=offset_and_message.message,
        envelope=envelope or Envelope(),
        force_payload_decoding=force_payload_decoding,
        reader_schema_id=reader_schema_id
    )


def _create_message_from_packed_message(
    packed_message,
    envelope,
    force_payload_decoding,
    kafka_position_info=None,
    reader_schema_id=None
):
    """ Builds a data_pipeline.message.Message from packed_message. If no
    reader schema id is provided, the schema used for encoding will be used for
    decoding.

    Args:
        packed_message (yelp_kafka.consumer.Message or kafka.common.KafkaMessage):
            The message info which has the payload, offset, partition,
            and key of the received message if of type yelp_kafka.consumer.message
            or just payload, uuid, schema_id in case of kafka.common.Message.
        force_payload_decoding (boolean): If this is set to `True` then
            we will decode the payload/previous_payload immediately.
            Otherwise the decoding will happen whenever the lazy *_data
            properties are accessed.
        kafka_position_info (Optional[KafkaPositionInfo]): The specified kafka
            position information.  The kafka_position_info may be constructed
            from the unpacked yelp_kafka message.
        reader_schema_id (Optional[int]): Schema id used to decode the incoming
            kafka message and build data_pipeline.message.Message message.
            Defaults to None.

    Returns (data_pipeline.message.Message):
        The message object
    """
    unpacked_message = envelope.unpack(packed_message.value)
    message_class = _message_type_to_class_map[unpacked_message['message_type']]
    message = message_class.create_from_unpacked_message(
        unpacked_message=unpacked_message,
        kafka_position_info=kafka_position_info,
        reader_schema_id=reader_schema_id
    )
    if force_payload_decoding:
        # Access the cached, but lazily-calculated, properties
        message.reload_data()
    return message



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from optparse import OptionGroup

from yelp_batch import Batch
from yelp_batch.batch import batch_command_line_options

from data_pipeline import __version__
from data_pipeline.schematizer_clientlib.models.refresh import Priority
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer
from data_pipeline.servlib.config_util import load_package_config


class FullRefreshJob(Batch):
    """
    FullRefreshJob parses command line arguments specifying full refresh jobs
    and registers the refresh jobs with the Schematizer.
    """

    def __init__(self):
        super(FullRefreshJob, self).__init__()
        self.notify_emails = ['bam+batch@yelp.com']

    @property
    def version(self):
        """Overriding this so we'll get the clientlib version number when
        the tailer is run with --version.
        """
        return "data_pipeline {}".format(__version__)

    @batch_command_line_options
    def define_options(self, option_parser):
        opt_group = OptionGroup(option_parser, 'Refresh Job Options')

        opt_group.add_option(
            '--source-id',
            dest='source_id',
            type='int',
            help='Source id of table to be refreshed.'
        )
        opt_group.add_option(
            '--source-name',
            type=str,
            help="Name of source within --namespace of table to be refreshed "
                 "(--namespace also required)"
        )
        opt_group.add_option(
            '--namespace',
            type=str,
            help="Name of namespace to narrow down sources to be refreshed "
                 "(--source-name also required)"
        )
        opt_group.add_option(
            '--offset',
            dest='offset',
            type='int',
            default=0,
            help='Row offset to start refreshing from. '
                 '(default: %default)'
        )
        opt_group.add_option(
            '--batch-size',
            dest='batch_size',
            type='int',
            default=500,
            help='Number of rows to process between commits '
                 '(default: %default).'
        )
        opt_group.add_option(
            '--priority',
            dest='priority',
            default='MEDIUM',
            help='Priority of this refresh: LOW, MEDIUM, HIGH or MAX '
                 '(default: %default)'
        )
        opt_group.add_option(
            '--filter-condition',
            dest='filter_condition',
            help='Custom WHERE clause to specify which rows to refresh '
                 'Note: This option takes everything that would come '
                 'after the WHERE in a sql statement. '
                 'e.g: --where="country=\'CA\' AND city=\'Waterloo\'"'
        )
        opt_group.add_option(
            '--avg-rows-per-second-cap',
            help='Caps the throughput per second. Important since without any control for this '
            'the batch can cause signifigant pipeline delays. (default: %default)',
            type='int',
            default=None
        )
        opt_group.add_option(
            '--config-path',
            dest='config_path',
            type='str',
            default='/nail/srv/configs/data_pipeline_tools.yaml',
            help='Config path for Refresh Job (default: %default)'
        )
        return opt_group

    def process_commandline_options(self, args=None):
        super(FullRefreshJob, self).process_commandline_options(args=args)
        if (self.options.avg_rows_per_second_cap is not None and
                self.options.avg_rows_per_second_cap <= 0):
            raise ValueError("--avg-rows-per-second-cap must be greater than 0")
        if self.options.batch_size <= 0:
            raise ValueError("--batch-size option must be greater than 0.")
        if not self.options.source_id and not (
            self.options.source_name and
            self.options.namespace
        ):
            raise ValueError("--source-id or both of--source-name and --namespace must be defined")
        if self.options.source_id and (
            self.options.source_name or
            self.options.namespace
        ):
            raise ValueError("Cannot use both --source-id and either of --namespace and --source-name")

        load_package_config(self.options.config_path)
        self.schematizer = get_schematizer()

    @property
    def source_id(self):
        if self.options.source_id:
            return self.options.source_id
        source_ids = [
            source.source_id
            for source in self.schematizer.get_sources_by_namespace(
                self.options.namespace
            ) if source.name == self.options.source_name
        ]
        if len(source_ids) == 1:
            return source_ids[0]
        elif len(source_ids) == 0:
            raise ValueError("Found no sources with namespace_name {} and source_name {}".format(
                self.options.namespace, self.options.source_name
            ))
        raise ValueError(
            "Pair of namespace_name {} and source_name {} somehow received more than one source. "
            "Investigation as to how is recommended.".format(
                self.options.namespace, self.options.source_name
            )
        )

    def run(self):
        self.job = self.schematizer.create_refresh(
            source_id=self.source_id,
            offset=self.options.offset,
            batch_size=self.options.batch_size,
            priority=Priority[self.options.priority],
            filter_condition=self.options.filter_condition,
            avg_rows_per_second_cap=self.options.avg_rows_per_second_cap
        )
        self.log.info(
            "Refresh registered with refresh id: {rid} "
            "on source id: {sid}".format(
                rid=self.job.refresh_id,
                sid=self.job.source.source_id
            )
        )


if __name__ == '__main__':
    FullRefreshJob().start()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import copy
import multiprocessing
import time
from collections import defaultdict

import simplejson as json
from cached_property import cached_property

from data_pipeline._kafka_producer import LoggingKafkaProducer
from data_pipeline._kafka_util import get_actual_published_messages_count
from data_pipeline._pooled_kafka_producer import PooledKafkaProducer
from data_pipeline.client import Client
from data_pipeline.config import get_config


logger = get_config().logger


class PublicationUnensurableError(Exception):
    pass


class Producer(Client):
    """Producers are context managers, that provide a high level interface for
    publishing :class:`data_pipeline.message.Message` instances into the data
    pipeline.

    When messages are handed to a producer via the :meth:`publish` method, they
    aren't immediately published into Kafka.  Instead, they're buffered until
    a number of messages are accumulated, or too much time has passed,
    then published all at once.  This process is designed to be largely
    transparent to the user.

    .. note::

        The clientlib used to include an AsyncProducer, which published to Kafka in the
        background.  This producer was somewhat flaky, increased development effort,
        and didn't provide a concrete performance benefit (see
        pb/150070 for benchmark results).  If we ever want to
        revive that producer, a SHA containing the producer just before its removal
        has been tagged as before-async-producer-removal.

    **Examples**:

      At it's simplest, start a producer and publish messages to it::

          with Producer() as producer:
              producer.publish(message)

      Messages are not immediately published, but are buffered.  Consequently,
      it may sometimes be necessary to flush the buffer before doing some
      tasks::

          with Producer() as producer:
              while upstream.has_another_batch_of_messages():
                  for message in upstream.get_messages_from_upstream():
                      producer.publish()

                  producer.flush()
                  upstream.all_those_messages_were published()

      The Producer is incapable of flushing its own buffers, which can be
      problematic if messages aren't published relatively constantly.  The
      :meth:`wake` should be called periodically to allow the Producer to clear
      its buffer if necessary, in the absence of messages.  If :meth:`wake`
      isn't called, messages in the buffer could be delayed indefinitely::

          with Producer() as producer:
              while True:
                  try:
                      message = slow_queue.get(block=True, timeout=0.1)
                      producer.publish(message)
                  except Empty:
                      producer.wake()

    Args:
      producer_name (str): See parameter `client_name` in
        :class:`data_pipeline.client.Client`.
      team_name (str): See parameter `team_name` in
        :class:`data_pipeline.client.Client`.
      expected_frequency_seconds (str): See parameter
        `expected_frequency_seconds` in :class:`data_pipeline.client.Client`.
      use_work_pool (bool): If true, the process will use a multiprocessing
        pool to serialize messages in preparation for transport.  The work pool
        can parallelize some expensive serialization.  Default is false.
      position_data_callback (Optional[function]): If provided, the function
        will be called when the producer starts, and whenever messages are
        committed to Kafka, with updated position data.  The callback should
        take a single argument, which will be an instance of
        :class:`PositionData`.
      dry_run (Optional[bool]): If true, producer will skip publishing message
        to kafka. Default is false.
      monitoring_enabled (Optional[bool]): If true, monitoring will be enabled
        to record client's activities. Default is true.
    """

    def __init__(
        self,
        producer_name,
        team_name,
        expected_frequency_seconds,
        use_work_pool=False,
        dry_run=False,
        position_data_callback=None,
        monitoring_enabled=True,
        schema_id_list=None
    ):
        super(Producer, self).__init__(
            producer_name,
            team_name,
            expected_frequency_seconds,
            monitoring_enabled,
            dry_run=dry_run
        )
        self.use_work_pool = use_work_pool
        self.dry_run = dry_run
        self.position_data_callback = position_data_callback
        if schema_id_list is None:
            schema_id_list = []
        # Send initial producer registration messages
        self.registrar.register_tracked_schema_ids(schema_id_list)

        self.enable_meteorite = get_config().enable_meteorite
        self.enable_sensu = get_config().enable_sensu
        self.monitors = {}
        self._next_sensu_update = 0
        self._sensu_window = 0
        self._setup_monitors()

    @cached_property
    def _kafka_producer(self):
        if self.use_work_pool:
            return PooledKafkaProducer(
                self._set_kafka_producer_position,
                dry_run=self.dry_run
            )
        else:
            return LoggingKafkaProducer(
                self._set_kafka_producer_position,
                dry_run=self.dry_run
            )

    @property
    def client_type(self):
        """String identifying the client type."""
        return 'producer'

    def __enter__(self):
        # By default, the kafka producer is created lazily, and doesn't
        # actually do anything until it needs to.  This method is used here to
        # force the kafka producer to wake up, which will guarantee that
        # its initialized before it is used.  This is important, since without
        # it, checkpoint position data won't be passed to the producer until
        # the user starts publishing messages.
        self.wake()
        self.registrar.start()

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        logger.info("Closing producer...")
        try:
            self.close()
            logger.info("Producer closed")
        except:
            logger.exception("Failed to close the Producer.")
            if exc_type is None:
                # The exception shouldn't mask the original exception if there
                # is one, but if an exception occurs, we want it to show up.
                raise
        # Returning any kind of truthy value will suppress the exception, if
        # there was one.  The intention of returning False here is to never
        # suppress the exception.  See:
        # https://docs.python.org/2/reference/datamodel.html#object.__exit__
        # for more information.
        return False

    def _setup_monitors(self):
        """This method sets up the meteorite monitor as well as the two sensu
        monitors, first for ttl, and second for delay.  The ttl monitor tracks
        the health of the producer and upstream heartbeat.  The delay monitor
        tracks whether the producer has fallen too far behind the upstream
        data"""

        try:
            from data_pipeline.tools.meteorite_wrappers import StatsCounter
            from data_pipeline.tools.sensu_alert_manager import SensuAlertManager
            from data_pipeline.tools.sensu_ttl_alerter import SensuTTLAlerter
        except ImportError:
            self.enable_meteorite = False
            self.enable_sensu = False
            return

        self.monitors["meteorite"] = StatsCounter(
            stat_counter_name=self.client_name,
            container_name=get_config().container_name,
            container_env=get_config().container_env
        )

        underscored_client_name = "_".join(self.client_name.split())
        # Sensu event dictionary parameters are described here:
        # http://pysensu-yelp.readthedocs.io/en/latest/index.html?highlight=send_event
        ttl_sensu_dict = {
            'name': "{0}_outage_check".format(underscored_client_name),
            'output': "{0} is back on track".format(self.client_name),
            'runbook': "y/datapipeline",
            'team': self.registrar.team_name,
            'page': get_config().sensu_page_on_critical,
            'status': 0,
            'ttl': "{0}s".format(get_config().sensu_ttl),
            'sensu_host': get_config().sensu_host,
            'source': "{0}_{1}".format(
                self.client_name,
                get_config().sensu_source
            ),
            'tip': "either the producer has died or there are no hearbeats upstream"
        }
        self._sensu_window = get_config().sensu_ping_window_seconds
        self.monitors["sensu_ttl"] = SensuTTLAlerter(
            sensu_event_info=ttl_sensu_dict,
            enable=self.enable_sensu
        )

        delay_sensu_dict = copy.deepcopy(ttl_sensu_dict)
        delay_sensu_dict.update({
            'name': "{0}_delay_check".format(underscored_client_name),
            'alert_after': get_config().sensu_alert_after_seconds,
        })
        disable_sensu = not self.enable_sensu
        SENSU_DELAY_ALERT_INTERVAL_SECONDS = 30
        self.monitors["sensu_delay"] = SensuAlertManager(
            SENSU_DELAY_ALERT_INTERVAL_SECONDS,
            self.client_name,
            delay_sensu_dict,
            get_config().max_producer_delay_seconds,
            disable=disable_sensu
        )

    def publish(self, message, timestamp=None):
        """Adds the message to the buffer to be published.  Messages are
        published after a number of messages are accumulated or after a
        slight time delay, whichever passes first.  Passing a message to
        publish does not guarantee that it will be successfully published into
        Kafka.

        **TODO(DATAPIPE-155|justinc)**:

        * Point to information about the message accumulation and time
          delay config.
        * Include information about checking which messages have actually
          been published.

        Args:
            message (data_pipeline.message.Message): message to publish
            timestamp (timezone aware timestamp): utc datetime of event
        """
        self._kafka_producer.publish(message)

        if self.enable_meteorite:
            self.monitors['meteorite'].process(message.topic)

        if self.enable_sensu and time.time() > self._next_sensu_update:
            self._next_sensu_update = time.time() + self._sensu_window
            self.monitors['sensu_ttl'].process()
            self.monitors['sensu_delay'].process(timestamp)

        self.monitor.record_message(message)
        self.registrar.update_schema_last_used_timestamp(
            message.schema_id,
            timestamp_in_milliseconds=long(1000 * time.time())
        )

    def ensure_messages_published(self, messages, topic_offsets):
        """This method should only be used when recovering after an unclean
        shutdown, and only if the upstream message source is persistent and can
        be rewound and replayed.  All messages produced since the last
        successful checkpoint should be passed as a list into this method,
        which will then ensure that each message has either already been
        published into Kafka, or will publish each message into Kafka.

        The call will block until all messages are published successfully.

        Immediately after calling this method, you should call
        :meth:`get_checkpoint_position_data` and persist the data.

        Args:
            messages (list of :class:`data_pipeline.message.Message`): List of
                messages to ensure are published.  The order of the messages
                matters, this code assumes that the messages are in the order
                they would have been published in.
            topic_offsets (dict of str to int): The topic offsets should be a
                dictionary containing the offset of the next message that would
                be published in each topic.  This should be in the format of
                :attr:`data_pipeline.position_data.PositionData.topic_to_kafka_offset_map`.

        Raises:
            PublicationUnensurableError: If any topics already have more messages
                published than would be published by this method, an assertion
                will fail.  This should never happen in practice.  If it
                does, it means that there has either been another publisher
                writing to the topic, which breaks the data pipeline contract,
                or there weren't enough messages passed into this method.  In
                either case, manual intervention will be required.  Note that
                in the event of a failure, some messages may have been
                published.
        """
        topic_messages_map = self._generate_topic_messages_map(messages)
        # raise_on_error must be set to False, otherwise this call will raise
        # an exception when any topic doesn't exist, preventing the topic from
        # ever being created in the context of ensure_messages_published
        topic_actual_published_count_map = (
            get_actual_published_messages_count(
                self._kafka_producer.kafka_client,
                topics=topic_messages_map.keys(),
                topic_tracked_offset_map=topic_offsets,
                raise_on_error=False
            )
        )

        # TODO(justinc|DATAPIPE-995): This can't be a set at this time, because
        # some data pipeline message fields aren't hashable - primarily the
        # upstream_position_info, which is a dict.
        already_published_messages = list()
        for topic, topic_messages in topic_messages_map.iteritems():
            # `get_actual_published_messages_count` only returns the message
            # count for topics that exist, so for non-existent topics, here it
            # sets the actual published message count to 0, i.e. high watermark
            # is 0.
            already_published_count = topic_actual_published_count_map.get(topic, 0)
            saved_offset = topic_offsets.get(topic, 0)

            info_to_log = dict(
                message="Attempting to ensure messages published",
                topic=topic,
                saved_offset=saved_offset,
                high_watermark=already_published_count + saved_offset,
                message_count=len(topic_messages),
                already_published_count=already_published_count
            )

            logger.info(json.dumps(info_to_log))

            if (
                already_published_count < 0 or
                already_published_count > len(topic_messages)
            ):
                # This is here primarily as a convenience to allow recovery
                # after logical errors.  It will result in breaking the
                # delivery guarantees.
                if get_config().force_recovery_from_publication_unensurable_error:
                    logger.critical(
                        "Forcing recovery from PublicationUnensurableError - "
                        "Intentionally Breaking Delivery Guarantees. "
                        "Turn force_recovery_from_publication_unensurable_error "
                        "off after recovery."
                    )
                    already_published_count = 0
                else:
                    raise PublicationUnensurableError()
            already_published_messages.extend(
                topic_messages[:already_published_count]
            )

        # Automatic flushing must be disabled while we're recovering, since
        # any partial flushing will result in state-saving callbacks being
        # triggered, which can cause downstream applications to update state
        # that really shouldn't be updated until all messages are published
        # successfully.
        position_tracker = self._kafka_producer.position_data_tracker
        with self._kafka_producer.disable_automatic_flushing():
            for message in messages:
                # We're recording already published messages here so that if
                # there's any ordering dependency related to state saving, we're
                # able to capture that.
                #
                # Concretely, imagine the last message has already been
                # published, and that messages come from a serial source like
                # db replication.  If we don't record the last message, even
                # though we're not actually publishing it, the state in the
                # producer will indicate the last message hasn't been published,
                # when we know that it has.  This breaks things if the
                # application crashes after this procedure, but before saving
                # again.
                if message in already_published_messages:
                    position_tracker.record_message(message)

                    topic_of_message = message.topic
                    already_published_count = topic_actual_published_count_map.get(
                        topic_of_message,
                        0
                    )
                    saved_offset = topic_offsets.get(topic_of_message, 0)
                    # This is required to update the high watermark for all the
                    # messages individually on the position tracker in-order to
                    # avoid offset in there from becoming stale.
                    position_tracker.update_high_watermark(
                        topic=message.topic,
                        offset=saved_offset,
                        message_count=already_published_count
                    )
                else:
                    self.publish(message)

            self.flush()

    def flush(self):
        """Block until all data pipeline messages have been
        successfully published into Kafka.
        """
        self._kafka_producer.flush_buffered_messages()

    def close(self):
        """Closes the producer, flushing all buffered messages into Kafka.
        Calling this method directly is not recommended, instead, use the
        producer as a context manager::

            with Producer() as producer:
                producer.publish(message)
                ...
                producer.publish(message)
        """
        self.registrar.stop()
        self.monitor.close()
        self._kafka_producer.close()
        assert len(multiprocessing.active_children()) == 0

    def get_checkpoint_position_data(self):
        """
        returns:
            PositionData: `PositionData` structure
                containing details about the last messages published to Kafka,
                including Kafka offsets and upstream position information.
        """
        return self.position_data

    def wake(self):
        """The synchronous producer has no mechanism to flush messages on its
        own, in the absence of other messages being published.  Consequently,
        if there are gaps where messages aren't published, this method should
        be called to allow the producer to flush its buffers if it needs to.

        If messages aren't published at least every 250ms, this method should
        be called about that often, to ensure that messages don't sit in the
        buffer for longer than that.

        Example::

            If the upstream messages are coming in slowly, or there can be gaps,
            call wake periodically so the producer has a change to publish
            messages::

                with Producer() as producer:
                    while True:
                        # if no new message arrive after 100ms, wake up the
                        # producer.
                        try:
                            message = slow_queue.get(block=True, timeout=0.1)
                            producer.publish(message)
                        except Empty:
                            producer.wake()
        """
        self._kafka_producer.wake()

    def _set_kafka_producer_position(self, position_data):
        """Called periodically to update the producer with position data.  This
        is expected to be called at least once when the KafkaProducer is started,
        and whenever messages are successfully published.

        Args:
            position_data (:class:PositionData): PositionData structure
                containing details about the last messages published to Kafka,
                including Kafka offsets and upstream position information.
        """
        self.position_data = position_data
        if self.position_data_callback:
            self.position_data_callback(position_data)

    def _generate_topic_messages_map(self, messages):
        topic_messages_map = defaultdict(list)
        for message in messages:
            topic_messages_map[message.topic].append(message)
        return topic_messages_map



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline.tools.heartbeat_periodic_processor import BasePeriodicProcessor
from data_pipeline.tools.meteorite_wrappers import StatGauge


class MeteoriteGaugeManager(BasePeriodicProcessor):
    """
    This class reports how far behind real-time the producer is to meteorite/signalfx

    Args:
       interval_in_seconds(int): the time interval between two events.
       stats_gauge_name(str): name of the stats gauge
       container_name(str): paasta container name
       container_env(str): paasta cluster name
       disable(bool): whether this gauge is disabled or not
       kwargs(dict): any additional keyword args for the Meteorite StatsGauge class
    """

    def __init__(
        self,
        interval_in_seconds,
        stats_gauge_name=None,
        container_name=None,
        container_env=None,
        disable=False,
        **kwargs
    ):
        super(MeteoriteGaugeManager, self).__init__(interval_in_seconds)
        self.gauge = StatGauge(
            stats_gauge_name,
            container_name=container_name,
            container_env=container_env,
            **kwargs
        )
        self.disable = disable

    def process(self, timestamp):
        if self.disable:
            return

        delay_seconds = (self._utc_now - timestamp).total_seconds()
        self.gauge.set(delay_seconds)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import time
from collections import defaultdict

import yelp_meteorite


class StatsBase(object):
    """ This class simply guarantees the other Stats classes will have a process method"""

    def process(self, topic):
        raise NotImplementedError


class StatsCounter(StatsBase):
    """ This class provides an easy way to send batched multi-dimension stats to meteorite/signalFX.
    Args:
      stats_counter_name(str): the name of this stat.
      message_counte_timer(float): the time interval between batch flushes.
      kwargs(dict): the stat dimensions
    """

    def __init__(self, stat_counter_name, message_count_timer=0.25, **kwargs):
        self.dimensions = kwargs

        self.message_count_timer = message_count_timer
        self._meteorite_counter = yelp_meteorite.create_counter(
            stat_counter_name,
            self.dimensions
        )
        self._reset()

    def _reset(self):
        self.counts = defaultdict(int)
        self.flush_time = time.time() + self.message_count_timer

    def increment(self, topic):
        """Increments the counter for the given topic"""
        self.counts[topic] += 1
        self.wake()

    def wake(self):
        """Allows the counter to flush to meteorite if necessary"""
        time_now = time.time()
        if time_now >= self.flush_time:
            self.flush()

    def flush(self):
        """Causes all of the existing counts to be sent to meteorite"""
        for topic, count in self.counts.iteritems():
            self._meteorite_counter.count(count, {'topic': topic})
        self._reset()

    def process(self, topic):
        """increments the counter for the topic specified

        Args:
            topic(str): the name of the topic to which a message is published
        """
        self.increment(topic)


class StatTimer(StatsBase):
    """This class exists primarily because we currently do not know if any sort
    of batching and/or sampling mechanisms will be necessary to avoid data loss
    in meteorite.
    """

    def __init__(self, stat_timer_name, **kwargs):
        self.dimensions = kwargs

        self._meteorite_timer = yelp_meteorite.create_timer(
            stat_timer_name,
            self.dimensions
        )

    def start(self):
        self._meteorite_timer.start()

    def stop(self, tmp_dimensions=None):
        self._meteorite_timer.stop(tmp_dimensions)


class StatGauge(StatsBase):
    """This class exists for the same reason as StatTimer."""

    def __init__(self, stat_gauge_name, **kwargs):
        self.dimensions = kwargs

        self._meteorite_gauge = yelp_meteorite.create_gauge(
            stat_gauge_name,
            self.dimensions
        )

    def set(self, value, tmp_dimensions=None):
        self._meteorite_gauge.set(value, tmp_dimensions)

    def process(self, value, tmp_dimensions):
        """sets the value of a gauge in meteorite

        Args:
            value(int): the value for the dimensions specified
            tmp_dimensions(dict): the dimensions for each emission
        """
        self.set(value, tmp_dimensions)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import defaultdict
from collections import Mapping

from data_pipeline.config import get_config
from data_pipeline.helpers.log import debug_log
from data_pipeline.position_data import PositionData


def PositionDataTracker():
    """Factory method for generating PositionDataTracker or subclasses
    """
    if get_config().merge_position_info_update:
        return _MergingPositionDataTracker()
    else:
        return _PositionDataTracker()


class _PositionDataTracker(object):
    """Makes it easier to construct and maintain PositionData.

    The basic idea is that messages are recorded as buffered, then recorded
    as published in matched pairs.  Position data can only be retrieved when
    the pairs actually match.
    """

    def __init__(self):
        self.unpublished_messages = 0
        self.topic_to_kafka_offset_map = {}
        self.merged_upstream_position_info_map = {}
        self._setup_position_info()

    def record_message(self, message):
        """In general, `record_message_buffered` should be preferred over
        `record_message`.  `record_message` is useful for recording messages
        that will never be published, but should still be recorded.  This use
        case is important for recovery procedures, that need to record already
        published messages alongside unpublished messages, so that state
        is saved correctly.  Don't call this method unless you're positive
        you need to.
        """
        if self._should_update_position_info(message):
            self._update_position_info(message)
        self._update_merged_upstream_position_info(message)

    def update_high_watermark(self, topic, offset, message_count):
        self.topic_to_kafka_offset_map[topic] = offset + message_count

    def record_message_buffered(self, message):
        debug_log(lambda: "Message buffered: %s" % repr(message))
        self.record_message(message)
        self.unpublished_messages += 1

    def record_messages_published(self, topic, offset, message_count):
        debug_log(
            lambda: "Messages published: %s, %s" % (topic, message_count)
        )
        self.update_high_watermark(topic, offset, message_count)
        self.unpublished_messages -= message_count

    def get_position_data(self):
        # Only allow checkpointing when there aren't unpublished messages
        assert self.unpublished_messages == 0
        return PositionData(
            last_published_message_position_info=self.last_published_message_position_info,
            topic_to_last_position_info_map=dict(self.topic_to_last_position_info_map),
            topic_to_kafka_offset_map=dict(self.topic_to_kafka_offset_map),
            merged_upstream_position_info_map=dict(self.merged_upstream_position_info_map)
        )

    def _setup_position_info(self):
        self.last_published_message_position_info = None
        self.topic_to_last_position_info_map = {}

    def _should_update_position_info(self, message):
        return (
            message.upstream_position_info is not None or
            not get_config().skip_position_info_update_when_not_set
        )

    def _update_position_info(self, message):
        self.last_published_message_position_info = message.upstream_position_info
        self.topic_to_last_position_info_map[message.topic] = message.upstream_position_info

    def _update_merged_upstream_position_info(self, message):
        if message.upstream_position_info is not None:
            _update_nested_dict(
                self.merged_upstream_position_info_map,
                message.upstream_position_info
            )


class _MergingPositionDataTracker(_PositionDataTracker):
    def _setup_position_info(self):
        self.last_published_message_position_info = {}
        self.topic_to_last_position_info_map = defaultdict(dict)

    def _update_position_info(self, message):
        self.last_published_message_position_info.update(
            message.upstream_position_info
        )
        self.topic_to_last_position_info_map[message.topic].update(
            message.upstream_position_info
        )


# TODO(joshszep|YELPLIB-65): Remove this and use function from yelp_lib
# when yelp-main makes it possible
def _update_nested_dict(original_dict, new_dict):
    """Update the dictionary and its nested dictionary fields.

    Note: This was copy-pasted from:
        opengrok/xref/submodules/yelp_lib/yelp_lib/containers/dicts.py?r=92297a46#40
        The reason is that this revision requires yelp_lib>=11.0.0 but we
        can not use this version yelp-main yet (see YELPLIB-65 for details).
        It's simpler to just temporarily pull this in.

    :param original_dict: Original dictionary
    :param new_dict: Dictionary with data to update
    """
    # Using our own stack to avoid recursion.
    stack = [(original_dict, new_dict)]
    while stack:
        original_dict, new_dict = stack.pop()
        for key, value in new_dict.items():
            if isinstance(value, Mapping):
                original_dict.setdefault(key, {})
                stack.append((original_dict[key], value))
            else:
                original_dict[key] = value



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
# flake8: noqa
"""
Load configuration for a service using `staticconf`_.

.. seealso::

    `staticconf overview`_ to get a better understanding of how
    staticconf works.


.. _staticconf: http://pythonhosted.org/PyStaticConfiguration
.. _staticconf overview: http://pythonhosted.org/PyStaticConfiguration/overview.html


Config Files
------------

Services should have at least one configuration file (usually ``config.yaml``).
This configuration file will be checked in and deployed with the service. It
should only include values which apply to all environments (dev, stage, prod).

The second configuration file is an optional file which contains values which
change for each environment (ex: dev, stage, prod). This file will be checked
in and deployed as part of the configs service repo
(ex: ``push@deploy-stagea:~/srv-configs``). During development this file is
often added to the service repo as ``config-env-dev.yaml``.

If both files are includes the contents of the `env config` are used as overrides
for the values in the default ``config.yaml``.  This is covered in more detail
below.


Loading
-------

Service configuration is loaded in two phases. First the contents of the config
file are loaded into the `DEFAULT` namespace. Anything in this config file will
be directly available using :mod:`staticconf.readers` and :mod:`staticconf.schema`
without specifying any namespace.

The second phase takes a list of `meta-config` from the `DEFAULT` namespace
(the default keys for this list are `module-config` and `module-env-config`),
and load each section of this list into the specified
:class:`staticconf.config.ConfigNamespace`. This second phase is
performed by :func:`configure_package`.

.. seealo:

    :func:`configure_packages` for details about the `meta-config` format


This module provides three functions to perform this loading process:


:func:`load_default_config`
    This function accepts the path to one or both config files. It performs
    both phases of `Loading`_ for one or both config files, using the
    default keys for the `meta-config` section. In most cases, this is the
    function you will use.

:func:`load_package_config`
    This function performs both phases of `Loading`_ for a single config file.


:func:`configure_packages`
    This is the lowest level function. It's responsible for taking a
    `meta-config` list and loading each section into a
    :class:`staticconf.config.ConfigNamespace`.


Examples
--------

This example illustrates the most common use of this module.

We start with a config file `config.yaml`:

.. code-block:: yaml

    # First the meta-config section
    module_config:

        - namespace: yelp_pyramid
          config:
              access_log_name: example_service_access_log
              error_log_name:  stream_example_service_error_log
              metrics_enabled: true

        - namespace: clog
          initialize: yelp_servlib.clog_util.initialize
          config:
              log_stream_name: example_service_app_log
          file:
              /nail/srv/configs/clog.yaml

        - namespace: services
          file: /nail/srv/configs/services.yaml

        - namespace: yelp_conn
          file: /nail/srv/configs/yelp_conn_generic.yaml
          config:
              connection_set_file: 'connection_sets.yaml'

        - namespace: gearman_topology
          file: /nail/srv/configs/gearman_topology.yaml

        - namespace: memcache_hosts
            file: /nail/srv/configs/memcache_hosts.yaml


    # And maybe some service specific configuration as well
    max_examples: 10
    paths_to_examples:
        - /nail/srv/example_service/foo.yaml
        - /nail/srv/example_service/bar.yaml


In our service code we can load this configuration. This loading will need to
happen in **each process**. So both the `webapp.py` and any other batches or
daemons will need to call this config loading.

.. code-block:: python

    from yelp_servlib.config_util import load_default_config

    # Load a configuration file using staticconf
    load_default_config(path_to_service_config_file)


Somewhere else in the service, you may read some of your configuration.

.. code-block:: python

    import staticconf

    paths_to_examples = staticconf.read_list('paths_to_examples')



Changes
-------

.. versionchanged:: 3.0.0
    :func:`initialize_clog` has been removed. It was deprecated in 2.x

.. versionchanged:: 3.0.0
    :func:`load_service_static_config` was removed. It was deprecated in 2.x

.. versionchanged:: 3.0.0
    :func:`read_log_level` was removed. It is available as
    :func:`staticconf.read_log_level`.



---------------------------


"""
from __future__ import absolute_import
from __future__ import unicode_literals

import logging
import os.path

import staticconf

log = logging.getLogger(__name__)


# Silence staticconf info to prevent unexpected config from entering logs
logging.getLogger('staticconf.config').setLevel(logging.WARN)


def load_package_config(filename, field='module_config', flatten=True):
    """Load the contents of a yaml configuration file using
    :func:`staticconf.loader.YamlConfiguration`, and use
    :func:`configure_packages` to load configuration at ``field``.

    Usage:

    .. code-block:: python

        from yelp_servlib import config_util

        config_util.load_package_config('config.yaml')


    :param filename: file path to a yaml config file with a `meta-config`
                     section
    :param field: field in the file `filename`, which contains a `meta-config`
                  section, which can be read by :func:`configure_packages`.
                  Defaults to `module_config`.
    :param flatten: boolean for whether staticconf should load each module
                  config with a flattened dictionary. Defaults to True.
    """
    config = staticconf.YamlConfiguration(filename)
    configs = config.get(field)
    if configs is None:
        log.warning('Field: {field} was not found in {filename}'.format(
            field=field,
            filename=filename
        ))
    else:
        configure_packages(configs, flatten=flatten)
    return config


def load_default_config(config_path, env_config_path=None, flatten=True):
    """Load service configuration assuming default config files and keys.

    This function will load configuration for one (`config.yaml`) or both
    (`config.yaml` and the env config) configuration files for a service.

    .. seealso::

        :func:`configure_packages` for expected config format of these files,
        and the operation performed on the `meta-config` section.

    .. seealso::

        `Config Files`_ and `Loading`_ for an overview.


    Usage:

    .. code-block:: python

        from yelp_servlib import config_util

        config_util.load_default_config('config.yaml',
                                        env_config_path='config-env-dev.yaml')


    :param config_path: the path to a yaml file with a `module_config` key
    :param env_config_path: the path to a yaml file with a
        `module_env_config` key. Defaults to `None`.
    :param flatten: boolean for whether staticconf should load each module
        config with a flattened dictionary. Defaults to True.
    """
    load_package_config(config_path, field='module_config', flatten=flatten)
    if env_config_path and os.path.exists(env_config_path):
        load_package_config(
            env_config_path, field='module_env_config', flatten=flatten)


def configure_packages(configs, ignore_initialize=False, flatten=True):
    """Load configuration into a :class:`staticconf.config.ConfigNamespace`
    and optionally call an initialize function after loading configuration
    for a package.


    `configs` should be a `meta-config`, which is a list of configuration
    sections.  Each section **must** have a `namespace` key and *may* have
    one or more other keys:

    namespace
        the name of a :class:`staticconf.config.ConfigNamespace` which will hold
        the configs loaded from a `file` or `config` in this section

    file
        the path to a yaml file. The contents of this file are loaded
        using :func:`staticconf.loader.YamlConfiguration`

    config
        a yaml mapping structure. The contents of this mapping are loaded
        using :func:`staticconf.loader.DictConfiguration`

    initialize
        a callable which accepts no arguments. It can be used to initialize
        state after configuration is loaded.

    For each section in the `meta-config` the following operations are
    performed:

        1. Load the contents of `file`, if the key is present
        2. Load the contents of `config`, if the key is present. If values
           were loaded from `file`, these new `config` values will override
           previous values.
        3. Call `initialize`, if the key is present

    Example configuration:

    .. code-block:: yaml

        module_config:
            -   namespace: yelp_pyramid
                config:
                    access_log_name: tmp_service_<service_name>
                    error_log_name: tmp_service_error_<service_name>
            -   namespace: clog
                initialize: yelp_servlib.clog_util.initialize
                config:
                    log_stream_name: <service_name>
                file:
                    /nail/srv/configs/clog.yaml
            -   namespace: yelp_memcache
                config:
                    clients:
                        encoders:
                            service_prefix: '<service_name>_service'

    Usage:

    .. code-block:: python

        configs = {...} # The contents of the example configuration above
        configure_packages(configs['module_config'])

    :param configs: List of config dicts.
    :param ignore_initialize: Whether to ignore the `initialize` key in the
                              config and not call it. This may be used to
                              reload a config whose initialize is only intended
                              to be called once. Defaults to `False`.
                              Available from version >=4.5.4.
    :param flatten: boolean for whether staticconf should load each module
        config with a flattened dictionary. Defaults to True.
    """
    for config in configs or []:
        # 1st load a yaml
        if 'file' in config:
            staticconf.YamlConfiguration(config['file'], namespace=config[
                                         'namespace'], flatten=flatten)

        # 2nd update with a config dict
        if 'config' in config:
            staticconf.DictConfiguration(config['config'], namespace=config[
                                         'namespace'], flatten=flatten)

        # 3rd call the initialize method
        if not ignore_initialize and 'initialize' in config:
            path = config['initialize'].split('.')
            function = path.pop()
            module_name = '.'.join(path)
            module = __import__(str(module_name), globals(), locals(), [str(path[-1])])
            getattr(module, function)()


initialize_submodules = configure_packages
"""Alias for :func:`configure_packages`"""



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline.tools.introspector.base_command import IntrospectorCommand


class SchemaCheckCommand(IntrospectorCommand):
    @classmethod
    def add_parser(cls, subparsers):
        schema_check_command_parser = subparsers.add_parser(
            "schema-check",
            description="Checks the compatibility of an avro schema and all"
                        " given avro_schemas within the given namespace"
                        " and source. Compatibility means that the schema can"
                        " deserialize data serialized by existing schemas within"
                        " all topics and vice-versa.",
            add_help=False
        )

        cls.add_base_arguments(schema_check_command_parser)
        cls.add_source_and_namespace_arguments(schema_check_command_parser)

        schema_check_command_parser.add_argument(
            "schema",
            type=str,
            help="The avro schema to check."
        )

        schema_check_command_parser.set_defaults(
            command=lambda args: cls("data_pipeline_instropsector_schema_check").run(
                args, schema_check_command_parser
            )
        )

    def process_args(self, args, parser):
        super(SchemaCheckCommand, self).process_args(args, parser)
        self.process_source_and_namespace_args(args, parser)
        self.schema = args.schema

    def is_compatible(self):
        is_compatible = self.schematizer.is_avro_schema_compatible(
            avro_schema_str=self.schema,
            source_name=self.source_name,
            namespace_name=self.namespace
        )
        return is_compatible

    def run(self, args, parser):
        self.process_args(args, parser)
        print {"is_compatible": self.is_compatible()}



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import signal
import sys

import kazoo.client
import yaml
from cached_property import cached_property
from cached_property import cached_property_with_ttl
from kazoo.exceptions import LockTimeout
from kazoo.retry import KazooRetry

from data_pipeline.config import get_config

log = get_config().logger

KAZOO_CLIENT_DEFAULTS = {
    'timeout': 30
}


class ZK(object):
    """A class for zookeeper interactions"""

    @property
    def max_tries(self):
        return 3

    @cached_property
    def ecosystem(self):
        return open(get_config().ecosystem_file_patch).read().strip()

    def __init__(self):
        retry_policy = KazooRetry(max_tries=self.max_tries)
        self.zk_client = self.get_kazoo_client(command_retry=retry_policy)
        self.zk_client.start()
        self.register_signal_handlers()

    @cached_property_with_ttl(ttl=2)
    def _local_zk(self):
        """Get (with caching) the local zookeeper cluster definition."""
        path = get_config().zookeeper_discovery_path.format(ecosystem=self.ecosystem)
        with open(path, 'r') as f:
            return yaml.safe_load(f)

    def _get_kazoo_client_for_cluster_def(self, cluster_def, **kwargs):
        """Get a KazooClient for a list of host-port pairs `cluster_def`."""
        host_string = ','.join('%s:%s' % (host, port) for host, port in cluster_def)

        for default_kwarg, default_value in KAZOO_CLIENT_DEFAULTS.iteritems():
            if default_kwarg not in kwargs:
                kwargs[default_kwarg] = default_value

        return kazoo.client.KazooClient(host_string, **kwargs)

    def get_kazoo_client(self, **kwargs):
        """Get a KazooClient for a local zookeeper cluster."""
        return self._get_kazoo_client_for_cluster_def(self._local_zk, **kwargs)

    def close(self):
        """Clean up the zookeeper client."""
        log.info("Stopping zookeeper")
        self.zk_client.stop()
        log.info("Closing zookeeper")
        self.zk_client.close()

    def _exit_gracefully(self, sig, frame):
        self.close()
        if sig == signal.SIGINT:
            self.original_int_handler(sig, frame)
        elif sig == signal.SITERM:
            self.original_term_handler(sig, frame)

    def register_signal_handlers(self):
        self.original_int_handler = signal.getsignal(signal.SIGINT)
        self.original_term_handler = signal.getsignal(signal.SIGTERM)
        signal.signal(signal.SIGINT, self._exit_gracefully)
        signal.signal(signal.SIGTERM, self._exit_gracefully)


class ZKLock(ZK):
    """Sets up zookeeper lock so that only one copy of the batch is run per cluster.
    This would make sure that data integrity is maintained (See DATAPIPE-309 for an example).
    Use it as a context manager (i.e. with ZKLock(name, namespace))."""

    def __init__(self, name, namespace, timeout=10):
        super(ZKLock, self).__init__()
        self.lock = self.zk_client.Lock("/{} - {}".format(name, namespace), namespace)
        self.timeout = timeout
        self.failed = False

    def __enter__(self):
        try:
            self.lock.acquire(timeout=self.timeout)
        except LockTimeout:
            log.warning("Already one instance running against this source! exit. See y/oneandonly for help.")
            self.close()
            sys.exit(1)

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        if self.lock.is_acquired:
            log.info("Releasing the lock...")
            self.lock.release()
        super(ZKLock, self).close()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import re


class DBSourcedNamespace(object):
    """Class that provides namespace information based on a given namespace name for database-source namespaces
    (i.e. ones created/used by replication handler).
    """
    ENVIRONMENTS = ['main', 'dev']

    def __init__(
        self,
        cluster,
        database,
        environment=None,
        suffixes=None
    ):
        """Params:
            cluster (str): The namespace's cluster (required if building from raw fields)
            database (str): The namespace's database (required if building from raw fields)
            environment (optional(str)): The namespace's environment
            suffixes (optional(listof(str))): The suffixes of the namespace (i.e. transformations applied to the namespace)
        """
        self.environment = environment
        self.cluster = cluster
        self.database = database
        self.suffixes = suffixes or []

    @classmethod
    def _build_from_sections(cls, sections, environment_exists):
        if environment_exists:
            cluster_pos = 1
            environment = sections[0]
        else:
            cluster_pos = 0
            environment = None
        cluster = sections[cluster_pos]
        database = sections[cluster_pos + 1]
        suffixes = sections[cluster_pos + 2:]
        return DBSourcedNamespace(cluster, database, environment, suffixes)

    @classmethod
    def create_from_namespace_name(cls, namespace_name):
        sections = namespace_name.split('.')
        cls._validate_sections(sections)
        return cls._build_from_sections(
            sections,
            cls._is_first_section_an_environment(sections)
        )

    @classmethod
    def create_from_namespace_name_with_guarantees(
        cls,
        namespace_name,
        expected_cluster=None,
        expected_database=None,
        expected_environment=None,
        expected_suffixes=None
    ):
        """For namespaces created by parsing names, it's possible that we want to modify how they are parsed based on
        something we know about the namespace in question to mitigate possible errors.
        e.g. for main.database.transformer where main is a cluster, normal parsing would treat it as an environment.

        Anything defined by the user will be forced on the name as true to parse the namespace name

        NOTE: Anything in suffixes will be guaranteed to be a subset of the end-suffixes
        """
        sections = namespace_name.split('.')
        cls._validate_sections(sections)
        namespace = cls._build_from_sections(
            sections,
            cls._is_first_section_an_environment_with_guarantees(
                sections,
                expected_cluster=expected_cluster,
                expected_database=expected_database,
                expected_environment=expected_environment,
                expected_suffixes=expected_suffixes
            )
        )
        cls.assert_expectations_satisfied(
            namespace,
            expected_cluster=expected_cluster,
            expected_database=expected_database,
            expected_environment=expected_environment,
            expected_suffixes=expected_suffixes
        )
        return namespace

    @classmethod
    def _is_first_section_an_environment(cls, sections):
        """Determines whether or not the first section is an environment or cluster definition"""
        return (
            sections[0] in cls.ENVIRONMENTS and
            # ex: main.database
            len(sections) >= 3
        )

    @classmethod
    def _is_first_section_an_environment_with_guarantees(
        cls,
        sections,
        expected_cluster=None,
        expected_database=None,
        expected_environment=None,
        expected_suffixes=None
    ):
        # We need to use all of our guarantees to determine if first section is environment
        # since we need all of these rules followed to maximally satisfy our guarantees
        # i.e.: We may only receive the suffixes ['transformer'] on a namespace of main.database.transformer,
        #       By making sure expected suffixes is followed,
        #       we will properly decide that main is not an environment but a cluster name.
        return sections[0] == expected_environment or (
            cls._is_first_section_an_environment(sections) and
            (not expected_cluster or expected_cluster == sections[1]) and
            (not expected_database or expected_database == sections[2]) and
            (not expected_suffixes or set(expected_suffixes).issubset(set(sections[3:])))
        )

    @classmethod
    def _validate_sections(cls, sections):
        if len(sections) < 2:
            raise ValueError(
                "not enough sections to split namespace_name to extract information "
                "(we need at least 2 to extract a cluster and database)"
            )
        for section in sections:
            if not re.match("^[_-]*[a-zA-Z0-9][a-zA-Z0-9_-]*$", section):
                raise ValueError(
                    "namespace_name section must contain at least one alphanumeric character and may include - or _"
                )

    @classmethod
    def assert_expectations_satisfied(
        cls,
        namespace,
        expected_cluster=None,
        expected_database=None,
        expected_environment=None,
        expected_suffixes=None
    ):
        if (
            (expected_environment and expected_environment != namespace.environment) or
            (expected_cluster and expected_cluster != namespace.cluster) or
            (expected_database and expected_database != namespace.database) or
            (expected_suffixes and not set(expected_suffixes).issubset(set(namespace.suffixes)))
        ):
            raise ValueError("Guarantees are impossible to rectify with existing namespace")

    def get_name(self):
        sections = []
        if self.environment:
            sections.append(self.environment)
        sections.append(self.cluster)
        sections.append(self.database)
        sections += self.suffixes
        return '.'.join(sections)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import signal
import sys
import time
from multiprocessing import Process
from optparse import OptionGroup

import psutil
from cached_property import cached_property
from yelp_batch import BatchDaemon
from yelp_batch.batch import batch_command_line_options
from yelp_batch.batch import batch_configure

from data_pipeline import __version__
from data_pipeline._namespace_util import DBSourcedNamespace
from data_pipeline.schematizer_clientlib.models.refresh import RefreshStatus
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer
from data_pipeline.servlib.config_util import load_package_config
from data_pipeline.tools.copy_table_to_blackhole_table import FullRefreshRunner
from data_pipeline.zookeeper import ZKLock

SCHEMATIZER_POLL_FREQUENCY_SECONDS = 5
COMPLETE_STATUSES = {RefreshStatus.SUCCESS, RefreshStatus.FAILED}


class FullRefreshManager(BatchDaemon):

    def __init__(self):
        super(FullRefreshManager, self).__init__()
        self.notify_emails = ['bam+batch@yelp.com']
        self.active_refresh = {'id': None, 'pid': None}

    @cached_property
    def schematizer(self):
        return get_schematizer()

    @property
    def version(self):
        """Overriding this so we'll get the clientlib version number when
        the tailer is run with --version.
        """
        return "data_pipeline {}".format(__version__)

    @batch_command_line_options
    def define_options(self, option_parser):
        opt_group = OptionGroup(option_parser, 'Full Refresh Manager Options')

        opt_group.add_option(
            '--namespace',
            type=str,
            default=None,
            help='Name of the namespace this refresh manager will handle. Expected format: '
            '"cluster.database"'
        )
        opt_group.add_option(
            '--config-path',
            dest='config_path',
            default='/nail/srv/configs/data_pipeline_tools.yaml',
            help='Config file path for the refresh manager. '
                 '(default: %default)'
        )
        opt_group.add_option(
            '--dry-run',
            action="store_true",
            default=False,
            dest="dry_run",
            help="Will execute all refreshes as dry runs, will still affect "
            "the schematizer's records. (default: %default)"
        )
        return opt_group

    @batch_configure
    def _init_global_state(self):
        if self.options.namespace is None:
            raise ValueError("--namespace is required to be defined")
        self.namespace = self.options.namespace
        self._set_cluster_and_database()
        self.config_path = self.options.config_path
        self.dry_run = self.options.dry_run
        load_package_config(self.config_path)
        # Removing the cmd line arguments to prevent child process error.
        sys.argv = sys.argv[:1]

    def _set_cluster_and_database(self):
        namespace_info = DBSourcedNamespace.create_from_namespace_name(self.namespace)
        self.cluster = namespace_info.cluster
        self.database = namespace_info.database

    def _begin_refresh_job(self, refresh):
        # We need to make 2 schematizer requests to get the primary_keys, but this happens infrequently enough
        # where it's not paricularly vital to make a new end point for it
        topic = self.schematizer.get_latest_topic_by_source_id(refresh.source.source_id)
        primary_keys = self.schematizer.get_latest_schema_by_topic_name(topic.name).primary_keys
        if len(primary_keys):
            primary_keys = ','.join(primary_keys)
        else:
            # Fallback in case we get bad data from the schematizer (command line default is 'id')
            primary_keys = 'id'
        refresh_batch = FullRefreshRunner(
            refresh_id=self.active_refresh['id'],
            cluster=self.cluster,
            database=self.database,
            config_path=self.config_path,
            table_name=refresh.source.name,
            offset=refresh.offset,
            batch_size=refresh.batch_size,
            primary=primary_keys,
            where_clause=refresh.filter_condition,
            dry_run=self.dry_run,
            avg_rows_per_second_cap=getattr(refresh, 'avg_rows_per_second_cap', None)
        )
        self.log.info(
            "Starting a batch table_name: {}, refresh_id: {}, worker_id: {}".format(
                refresh.source.name,
                self.active_refresh['id'],
                self.active_refresh['pid']
            )
        )
        refresh_batch.start()

    def setup_new_refresh(self, refresh):
        active_pid = self.active_refresh['pid']
        if active_pid is not None:
            # If we somehow have an active worker but not an active refresh (refresh died somehow),
            # then we want to kill that worker.
            os.kill(active_pid, signal.SIGTERM)
        new_worker = Process(
            target=self._begin_refresh_job,
            args=(refresh,)
        )
        self.active_refresh['id'] = refresh.refresh_id
        new_worker.start()
        self.active_refresh['pid'] = new_worker.pid

    def _should_run_next_refresh(self, next_refresh):
        if not self.active_refresh['id']:
            return True
        current_refresh = self.schematizer.get_refresh_by_id(
            self.active_refresh['id']
        )
        next_priority = next_refresh.priority.value
        current_priority = current_refresh.priority.value
        return (next_priority > current_priority or
                current_refresh.status in COMPLETE_STATUSES)

    def determine_best_refresh(self, not_started_jobs, paused_jobs):
        if not_started_jobs and paused_jobs:
            not_started_job = not_started_jobs[0]
            paused_job = paused_jobs[0]
            if not_started_job.priority.value > paused_job.priority.value:
                return not_started_job
            else:
                return paused_job

        if not_started_jobs:
            return not_started_jobs[0]

        if paused_jobs:
            return paused_jobs[0]

        return None

    def set_zombie_refresh_to_fail(self):
        """The manager sometimes gets in to situations where
        the worker becomes a zombie but the refresh stays 'IN_PROGRESS'.
        For these situations we want to correct the refresh status to failed.
        """
        current_pid = self.active_refresh['pid']
        if current_pid is None:
            return

        p = psutil.Process(current_pid)
        if p.status() != psutil.STATUS_ZOMBIE:
            return

        refresh = self.schematizer.get_refresh_by_id(
            self.active_refresh['id']
        )
        if refresh.status.value == RefreshStatus.IN_PROGRESS.value:
            self.schematizer.update_refresh(
                self.active_refresh['id'],
                RefreshStatus.FAILED,
                0
            )
            self.active_refresh['id'] = None
            self.active_refresh['pid'] = None

    def get_next_refresh(self):
        not_started_jobs = self.schematizer.get_refreshes_by_criteria(
            self.namespace,
            RefreshStatus.NOT_STARTED
        )
        paused_jobs = self.schematizer.get_refreshes_by_criteria(
            self.namespace,
            RefreshStatus.PAUSED
        )
        return self.determine_best_refresh(not_started_jobs, paused_jobs)

    def run(self):
        with ZKLock(name="refresh_manager", namespace=self.namespace):
            try:
                while True:
                    self.set_zombie_refresh_to_fail()
                    next_refresh = self.get_next_refresh()
                    if next_refresh and self._should_run_next_refresh(next_refresh):
                        self.setup_new_refresh(next_refresh)
                    if self._stopping:
                        break
                    time.sleep(SCHEMATIZER_POLL_FREQUENCY_SECONDS)
            finally:
                if self.active_refresh['pid'] is not None:
                    # When the manager goes down, the active refresh is paused.
                    os.kill(self.active_refresh['pid'], signal.SIGTERM)


if __name__ == '__main__':
    FullRefreshManager().start()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline.tools.introspector.register.base_command import _BaseRegisterCommand


class RegisterMysqlCommand(_BaseRegisterCommand):
    @classmethod
    def add_parser(cls, subparsers):
        register_mysql_command_parser = subparsers.add_parser(
            "mysql",
            description="Register the given mysql statements "
                        "as schemas to the schematizer.",
            add_help=False
        )

        cls.add_base_arguments(register_mysql_command_parser)

        register_mysql_command_parser.add_argument(
            "--create-table",
            type=str,
            required=True,
            help="The mysql statement of creating new table"
        )

        register_mysql_command_parser.add_argument(
            "--old-create-table",
            type=str,
            default=None,
            help="The mysql statement of creating old table. "
        )

        register_mysql_command_parser.add_argument(
            "--alter-table",
            type=str,
            default=None,
            help="The mysql statement of altering table schema. "
        )

        register_mysql_command_parser.set_defaults(
            command=lambda args: cls("data_pipeline_instropsector_register_mysql").run(
                args, register_mysql_command_parser
            )
        )

    def process_args(self, args, parser):
        super(RegisterMysqlCommand, self).process_args(args, parser)
        self.create_table = args.create_table
        self.old_create_table = args.old_create_table
        self.alter_table = args.alter_table

    def run(self, args, parser):
        self.process_args(args, parser)
        schema = self.schematizer.register_schema_from_mysql_stmts(
            namespace=self.namespace,
            source=self.source_name,
            source_owner_email=self.source_owner_email,
            contains_pii=self.pii,
            new_create_table_stmt=self.create_table,
            old_create_table_stmt=self.old_create_table,
            alter_table_stmt=self.alter_table
        )
        self.print_schema(schema)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from enum import Enum


class _EnumRepr(Enum):
    def __repr__(self):
        return '{}.{}({})'.format(
            self.__class__.__name__,
            self.name,
            self.value
        )


class MessageType(_EnumRepr):
    """Messages should be published primarily using the create, update, and
    delete types.  Refresh messages should only be produced if you know what
    you're doing, if in doubt, ask please.

    Attributes:
      create: when new data is created, the payload contains the contents of
        the new row
      update: when data is updated, payload contains the new content and
        previous_payload contains the old content of the row
      delete: when rows are removed, the payload contains the content of the
        row before removal
      refresh: refresh messages are used to intially populate a topic, they
        do not correspond to any particular data change
    """
    log = 0
    create = 1
    update = 2
    delete = 3
    refresh = 4


class _ProtectedMessageType(_EnumRepr):
    """Protected message types should generally be avoided.  The clientlib
    won't expose these messages to users, they're used internally only.

    Attributes:
      heartbeat: emitted periodically on low volume topics so auditing
        processes can differentiate between slow or stalled topics and
        topics without messages.
      monitor: monitor messages are used to count the number of messages
        produced/consumed by client in a given time frame
    """
    heartbeat = 5
    monitor = 6
    registration = 7



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import simplejson

from data_pipeline.tools.introspector.base_command import IntrospectorCommand
from data_pipeline.tools.introspector.models import IntrospectorNamespace


class NamespaceInfoCommand(IntrospectorCommand):
    @classmethod
    def add_parser(cls, subparsers):
        info_command_parser = subparsers.add_parser(
            "namespace",
            description="Get information on a specific data pipeline namespace.",
            add_help=False
        )

        info_command_parser.add_argument(
            '--active-namespaces',
            default=False,
            action='store_true',
            help=(
                'If set, this command will also return information about active '
                'sources and topics for this namespace. '
                'This is a time expensive operation.'
            )
        )

        cls.add_base_arguments(info_command_parser)

        info_command_parser.add_argument(
            "namespace_name",
            type=str,
            help="Name of namespace to retrieve information on."
        )

        info_command_parser.set_defaults(
            command=lambda args: cls("data_pipeline_instropsector_info_namespace").run(
                args,
                info_command_parser
            )
        )

    def info_namespace(self, name, active_namespaces=False):
        namespaces = self.schematizer.get_namespaces()
        info_namespace = None
        for namespace in namespaces:
            if namespace.name == name:
                info_namespace = namespace
                break
        if info_namespace:
            namespace = IntrospectorNamespace(
                namespace,
                active_namespaces=(self.active_namespaces if active_namespaces else None)
            ).to_ordered_dict()
            namespace['sources'] = self.list_sources(
                namespace_name=namespace['name']
            )
            return namespace
        else:
            raise ValueError("Given namespace doesn't exist")

    def process_args(self, args, parser):
        super(NamespaceInfoCommand, self).process_args(args, parser)
        self.namespace_name = args.namespace_name

    def run(self, args, parser):
        self.process_args(args, parser)
        print simplejson.dumps(
            self.info_namespace(
                self.namespace_name,
                active_namespaces=args.active_namespaces
            )
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import simplejson as json
from cached_property import cached_property


class SchemaRef(object):
    """ Object representation of the JSON specification for a schema reference.
    A schema reference contains metadata such as notes, docs, categories, and
    other attributes useful to the Documentation Tool.

    The JSON specification is as follows:
        {
          'doc_source': '', // A string, usually a URL, of the original
                            // source the schema ref was generated from.

          'doc_owner': '',  // This will be the username for 'last updated by'
                            // in all notes - it should be a valid email,
                            // usually a team email such as 'bam@yelp.com'

          'docs': [         // An array of dicts, one for each table
            {
              'namespace': '',  // The namespace of the table, which should
                                // be the DB schema if generating docs
                                // for existing tables.

              'source': '', // The source name, which should be the DB table
                            // if generating docs for existing tables.

              'doc': '',    // This will fill the 'description' for the table

              'note': '',   // This will fill the 'notes' for the table

              'category': '',   // This will fill the 'category' for the table

              'owner_email': '',    // This will fill the 'owner' for the table

              'file_display': '',   // This is used for display name of the
                                    // source_file in the documentation tool.
                                    // Generally should be the path to the model
                                    // in relation to the base of the repo

              'file_url': '',   // URL of the file in opengrok/gitweb - used
                                // for the 'source_file' hyperlink target in the
                                // documentation tool

              'contains_pii': false,    // Boolean true/false of if the table
                                        // contains pii

              'fields': [   // An array of dicts, one for each column
                {
                         'name': '',    // Column name of the field
                         'doc': '',     // 'Description' for the column
                         'note': ''     // 'Notes' for the column
                }
              ]
            },
          ]
        }

    See DATAPIPE-259 for more information.
    """

    @classmethod
    def load_from_file(cls, schema_ref_path, **kwargs):
        """ If a schema ref is specified, load, parse, and return it. Otherwise
        return a SchemaRef built from an empty dictionary.
        """
        schema_ref = {}
        if schema_ref_path:
            with open(schema_ref_path) as schema_ref_file:
                schema_ref = json.load(schema_ref_file)
        return cls(schema_ref=schema_ref, **kwargs)

    def __init__(
            self,
            schema_ref,
            defaults
    ):
        """ Construct a SchemaRef from a parsed schema_ref json object.

        Args:
            schema_ref (dict): The parsed schema_ref json object.
            defaults (dict): A key-value pair of schema_ref attributes to
                default values. The following keys are suggested:
                        'doc_owner'    # string, email of doc owner
                        'owner_email'  # string, email of source owner
                        'namespace'    # string, default namespace
                        'doc'          # string, default docstring
                        'contains_pii' # bool, default contains_pii flag
                        'category'     # string, default category
        """
        self.schema_ref = schema_ref
        self.defaults = defaults

    @cached_property
    def _source_to_ref_map(self):
        """ A dictionary that maps source name to schema ref node. If
        no schema ref is loaded, this will return an empty dictionary.
        """
        return {ref['source']: ref for ref in self.schema_ref.get('docs', [])}

    def get_source_ref(self, source):
        """ Get the source ref for a given source name, or None if there
        is none present.
        """
        return self._source_to_ref_map.get(source)

    @cached_property
    def doc_owner(self):
        """ The documentation owner as defined by the schema_ref, falling back
        to the specified default if none was provided.
        """
        return self.get_ref_val(self.schema_ref, 'doc_owner')

    def get_source_val(self, source, key):
        """ Get a value for a given key in the given source name, returning
        defaults where required. For example:
        ``schema_ref.get_source_val('business_dimension', 'note')`` will
        return the 'note' value for the object in the 'docs' list with a
        'source' == 'business_dimension', or the default note if none is
        available.
        """
        return self.get_ref_val(self.get_source_ref(source), key)

    def get_ref_val(self, ref, key):
        """ Get a value for a given key in the given ref object. A ref object
        is a node in the schema_ref and should either be the top-level ref (aka
        `self.schema_ref`), a source ref (an object in the 'docs' list), or
        a field ref (an object in a source ref's 'fields' list.) For example:
        ``schema_ref.get_ref_val(field_ref, 'note')`` will return the 'note'
        value for the field_ref object, or the default note if none is
        available.
        """
        if ref:
            return ref.get(key, self.defaults.get(key))
        else:
            return self.defaults.get(key)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline_avro_util.avro_string_reader import AvroStringReader
from data_pipeline_avro_util.avro_string_writer import AvroStringWriter

from data_pipeline.helpers.singleton import Singleton
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer


class _AvroStringStore(object):
    """Singleton instance of store that caches
    AvroStringsWriter and AvroStringReader objects perticularly
    used by message class to encode and decode messages respectively.

    This class was added for performance enhancements
    w store : pb/199453
    w/o store : pb/199448
    """
    __metaclass__ = Singleton

    def __init__(self):
        self._writer_cache = {}
        self._reader_cache = {}

    @property
    def _schematizer(self):
        return get_schematizer()

    def _get_avro_schema(self, schema_id):
        return self._schematizer.get_schema_by_id(
            schema_id
        ).schema_json

    def get_writer(self, id_key, avro_schema=None):
        key = id_key
        avro_string_writer = self._writer_cache.get(key)
        if avro_string_writer:
            return avro_string_writer

        avro_schema = avro_schema or self._get_avro_schema(id_key)
        avro_string_writer = AvroStringWriter(schema=avro_schema)
        self._writer_cache[key] = avro_string_writer
        return avro_string_writer

    def get_reader(
        self,
        reader_id_key,
        writer_id_key,
        reader_avro_schema=None,
        writer_avro_schema=None
    ):
        key = reader_id_key, writer_id_key
        avro_string_reader = self._reader_cache.get(key)
        if avro_string_reader:
            return avro_string_reader

        reader_schema = (
            reader_avro_schema or self._get_avro_schema(reader_id_key)
        )
        writer_schema = (
            writer_avro_schema or self._get_avro_schema(writer_id_key)
        )
        avro_string_reader = AvroStringReader(
            reader_schema=reader_schema,
            writer_schema=writer_schema
        )
        self._reader_cache[key] = avro_string_reader
        return avro_string_reader



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Define retry policies, retry functions on specified conditions or exceptions.
It doesn't have decorators for retry functions yet but they can be added later.
TODO: [DATAPIPE-399|clin] add retry decorators.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import random
import time
from collections import namedtuple

from data_pipeline.config import get_config


logger = get_config().logger

UNLIMITED = -1


def calc_next_exponential_backoff_delay(
    current_delay_secs,
    backoff_factor,
    max_delay_secs,
    with_jitter=False
):
    """Calculate the delay (in seconds) before next retry based on
    current delay using exponential backoff, with optional jitter.

    See http://www.awsarchitectureblog.com/2015/03/backoff.html for
    information about exponential backoff and jitter.

    Args:
        current_delay_secs (float): current backoff delay in seconds.
        backoff_factor (int): the constant used for exponential backoff.
        max_delay_secs (float): maximum backoff delay in seconds.
        with_jitter(Optional[bool]): whether to use jitter.  Default to False.

    Returns (float):
        Next backoff delay in seconds.
    """
    next_delay_secs = min(
        max_delay_secs,
        current_delay_secs * backoff_factor
    )
    if with_jitter:
        next_delay_secs = random.uniform(0, next_delay_secs)
    return next_delay_secs


class MaxRetryError(Exception):
    """Exception raised when the target no longer can be retried and
    it hasn't been successfully finished.
    """

    def __init__(self, last_result=None, *args):
        super(MaxRetryError, self).__init__(*args)
        self._last_result = last_result

    @property
    def last_result(self):
        """The result of the target function from the last retry before
        the max retry criteria is reached.  It could be an exception or
        the return value of the target function.
        """
        return self._last_result


class Predicate(object):
    """Predicate function and its arguments.  A predicate is a function
    that returns a boolean.

    Args:
        predicate (function): A function that returns a boolean.
        predicate_params (dict): keyword parameters of predicate function.
    """

    def __init__(self, predicate, **kwargs):
        self.predicate_func = predicate
        self.predicate_params = kwargs

    @property
    def is_true(self):
        return self.predicate_func(**self.predicate_params)


class BackoffPolicy(object):
    """Base class to define each specific backoff policy.
    """

    def next_backoff_delay(self):
        raise NotImplementedError()


class ConstantBackoffPolicy(BackoffPolicy):
    """Constant backoff waiting time between retries.

    Args:
        delay_seconds (Optional[float]): Seconds to wait between retries.
            Default to 1 second.
    """

    def __init__(self, delay_seconds=1):
        super(ConstantBackoffPolicy, self).__init__()
        self.delay_seconds = delay_seconds

    def next_backoff_delay(self):
        return self.delay_seconds


class ExpBackoffPolicy(BackoffPolicy):
    """Exponential backoff with optional jitter.

    Args:
        initial_delay_secs (Optional[float]): Seconds to wait before the first
            retry.  Default to 1 second.
        max_delay_secs (Optional[float]): Maximum wait time in seconds between
            retries.  Default to 3 minutes.
        backoff_factor (Optional[int]): constant used to calculate exponential
            backoff delay.  Default to 2.
        with_jitter (Optional[bool]): Whether to introduce randomness when
            calculating backoff delay.  Default to False.
    """

    def __init__(
        self,
        initial_delay_secs=1,
        max_delay_secs=3 * 60,
        backoff_factor=2,
        with_jitter=False
    ):
        super(ExpBackoffPolicy, self).__init__()
        self.initial_delay_secs = initial_delay_secs
        self.max_delay_secs = max_delay_secs
        self.backoff_factor = backoff_factor
        self.with_jitter = with_jitter
        self._current_delay_secs = self.initial_delay_secs

    def next_backoff_delay(self):
        """Calculate the next backoff delay in seconds based on the current
        backoff delay time.
        """
        next_delay = calc_next_exponential_backoff_delay(
            current_delay_secs=self._current_delay_secs,
            max_delay_secs=self.max_delay_secs,
            backoff_factor=self.backoff_factor,
            with_jitter=self.with_jitter
        )
        self._current_delay_secs = next_delay
        return next_delay


"""Policy that defines how to retry and retry cap.
    Args:
        backoff_policy (:class:`_retry_util.BackoffPolicy`): One of the defined
            backoff policy, such as :class:`_retry_util.ExpBackoffPolicy`.
        max_retry_count (Optional[int]): Maximum number of retries.  Default
            to unlimited.  Retry stops if max_retry_count or max_runtime_secs
            is exceeded, whichever comes first.
        max_runtime_secs(Optional[float]): Maximum elapsed time the target
            can keep retrying.  Default to unlimited, i.e. retrying the target
            indefinitely.  Retry stops if max_retry_count or max_runtime_secs
            is exceeded, whichever comes first.
"""
RetryPolicy = namedtuple(
    'RetryPolicy',
    'backoff_policy, max_retry_count, max_runtime_secs'
)
RetryPolicy.__new__.__defaults__ = (UNLIMITED, UNLIMITED)


def retry_on_condition(retry_policy, retry_conditions,
                       func_to_retry, use_previous_result_as_param=False,
                       *args, **kwargs):
    """Retry the given function in the given retry conditions with specified
    retry policy.

    Args:
        retry_policy (:class:`_retry_util.RetryPolicy`): Retry policy that
            specifies how to retry the target function.
        retry_conditions ([:class:`_retry_util.Predicate`]): List of conditions
            in which the function should be retried.  Each condition is a
            predicate which returns True (should retry) or False.
        func_to_retry (function): function that needs to be retried.
        args (list): non-keyword arguments of `func_to_retry`.
        kwargs (dict): keyword arguments of `func_to_retry`.
        use_previous_result_as_param (Optional[bool]): Whether to use the
            result of `func_to_retry` as new parameters in the next retry.
            Default to False.  When using this feature, one limitation is
            `func_to_retry` must return a single value, a tuple, or a dict,
            which contains the new values of all the function parameters.

            For example, to retry function `triple`:
                retry_on_conditions(
                    retry_policy,
                    retry_conditions,
                    func_to_retry=triple,
                    user_previous_result_as_param=True,
                    i=10
                )
            function `triple` can be defined as:
                def triple(i):
                    return i * 3  # single value
            or
                def triple(i):
                    return {'i': i * 3}  # keyword parameters
            or
                def triple(i):
                    return (i,)  # non-keyword parameters

    Returns:
        The return value of `func_to_retry` if it successfully finishes.
        It raises :class:`_retry_util.MaxRetryError` exception if the
        function still fails and no longer can be retried.
    """
    retry_tracker = _RetryTracker(
        retry_policy.max_retry_count,
        retry_policy.max_runtime_secs
    )
    retry_tracker.start()
    while True:
        result = func_to_retry(*args, **kwargs)
        if all(not predicate.is_true for predicate in retry_conditions):
            return result

        if not retry_tracker.exceeded_max_retry():
            if use_previous_result_as_param:
                args, kwargs = _get_func_params_from_result(result)
            retry_tracker.increment_retry_count()
            time.sleep(retry_policy.backoff_policy.next_backoff_delay())
            continue
        raise MaxRetryError(last_result=result)


def _get_func_params_from_result(result):
    if isinstance(result, (list, tuple)):
        return tuple(result), {}
    if isinstance(result, dict):
        return (), result
    return tuple([result]), {}


def retry_on_exception(retry_policy, retry_exceptions,
                       func_to_retry, *args, **kwargs):
    """Retry the given function with specified retry policy when one of
    specified exceptions occurs.

    Args:
        retry_policy (:class:`_retry_util.RetryPolicy`): Retry policy that
            specifies how to retry the target function.
        retry_exceptions ([:class:`Exception`]): List of exceptions the target
            function should be retried when one of which occurs.
        func_to_retry (function): function that needs to be retried.
        args (list): non-keyword arguments of `func_to_retry`.
        kwargs (dict): keyword arguments of `func_to_retry`.

    Returns:
        The return value of `func_to_retry` if it successfully finishes.
        It raises :class:`_retry_util.MaxRetryError` exception if the
        function still fails and no longer can be retried.
    """
    retry_tracker = _RetryTracker(
        retry_policy.max_retry_count,
        retry_policy.max_runtime_secs
    )
    retry_tracker.start()
    while True:
        try:
            return func_to_retry(*args, **kwargs)
        except retry_exceptions as e:
            if not retry_tracker.exceeded_max_retry():
                retry_tracker.increment_retry_count()
                time.sleep(retry_policy.backoff_policy.next_backoff_delay())
                continue
            raise MaxRetryError(last_result=e)


class _RetryTracker(object):

    def __init__(self, max_retry_count, max_runtime_secs):
        self.max_retry_count = max_retry_count
        self.max_runtime_secs = max_runtime_secs
        self._retried_count = 0
        self._start_runtime = None

    def start(self):
        self._start_runtime = time.time()
        self._retried_count = 0

    def increment_retry_count(self, count=1):
        self._retried_count += count

    def exceeded_max_retry(self):
        return self._exceeded_max_retry_count() or self._exceeded_max_runtime()

    def _exceeded_max_retry_count(self):
        return (self.max_retry_count != UNLIMITED and
                self.max_retry_count <= self.retried_count)

    def _exceeded_max_runtime(self):
        return (self.max_runtime_secs != UNLIMITED and
                self.max_runtime_secs <= self.elapsed_runtime_secs)

    @property
    def retried_count(self):
        """Number of retry attempts so far.
        """
        return self._retried_count

    @property
    def elapsed_runtime_secs(self):
        """Total seconds elapsed since tracker starts.
        """
        return (time.time() - self._start_runtime).total_seconds()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

from data_pipeline.schematizer_clientlib.models.model_base import BaseModel
from data_pipeline.schematizer_clientlib.models.note import _Note


"""
Represent the data of an Avro schema element.

Args:
    id (int): The element id.
    schema_id (int): The id of the avro schema.
    element_type (): The data type of the element
    element_name (str): The column corresponding to the key of the AvroSchemaElement
    doc ():
    note (Optional[data_pipeline.schematizer_clientlib.models.note.Note]): Information specified by users about the schema.
    created_at (str): The timestamp when the schema is created in ISO-8601
        format.
    updated_at (str): The timestamp when the schema is last updated in ISO-8601
        format.
"""

AvroSchemaElement = namedtuple(
    'AvroSchemaElement',
    ['id', 'schema_id', 'element_type', 'element_name', 'doc',
     'note', 'created_at', 'updated_at']
)

_SCHEMA_KEY_DELIMITER = '|'


class _AvroSchemaElement(BaseModel):
    """Internal class used to convert from/to various data structure and
    facilitate constructing the return value of schematizer functions.
    """

    def __init__(self, id, schema_id, element_type, key, doc, note,
                 created_at, updated_at):
        self.id = id
        self.schema_id = schema_id
        self.element_type = element_type
        self.element_name = None
        self.doc = doc
        self.note = note
        self.created_at = created_at
        self.updated_at = updated_at
        split_keys = key.split(_SCHEMA_KEY_DELIMITER)
        if len(split_keys) >= 2:
            self.element_name = split_keys[1]

    @classmethod
    def from_response(cls, response_lst):
        res = []
        for response in response_lst:
            res.append(
                cls(
                    id=response.id,
                    schema_id=response.schema_id,
                    element_type=response.element_type,
                    key=response.key,
                    doc=response.doc,
                    note=_Note.from_response(response.note),
                    created_at=response.created_at,
                    updated_at=response.updated_at
                )
            )
        return res

    def to_result(self):
        return AvroSchemaElement(
            id=self.id,
            schema_id=self.schema_id,
            element_type=self.element_type,
            element_name=self.element_name,
            doc=self.doc,
            note=self.note.to_result() if self.note is not None else None,
            created_at=self.created_at,
            updated_at=self.updated_at
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This module contains classes that implement retry logic that provides various
publish guarantees.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

from kafka.common import LeaderNotAvailableError

from data_pipeline._kafka_util import get_actual_published_messages_count
from data_pipeline.config import get_config
from data_pipeline.publish_guarantee import PublishGuaranteeEnum


logger = get_config().logger


_TopicPartition = namedtuple('_TopicPartition', ['topic_name', 'partition'])


_Stats = namedtuple('_Stats', ['original_offset', 'message_count'])


class RetryHandler(object):
    """The class tracks the message publishing statistics in each retry,
    such as topic offset, number of published messages, etc., and determines
    which messages should be retried based on specified publishing guarantee.
    """

    # The logging is explicitly added in the class instead of being delegated to
    # `LoggingKafkaProducer` to help tracking issues when they occur.

    def __init__(
        self,
        kafka_client,
        requests,
        publish_guarantee=PublishGuaranteeEnum.exact_once,
    ):
        self.kafka_client = kafka_client
        self.initial_requests = requests
        self.requests_to_be_sent = requests
        self.publish_guarantee = publish_guarantee
        self.success_topic_stats_map = {}
        self.success_topic_accum_stats_map = {}

    def update_requests_to_be_sent(self, responses, topic_offsets=None):
        """Update stats from the responses of the publishing requests and
        determine which messages should be retried.

        Args:
            responses (kafka.common.FetchResponse or kafka.common.KafkaError):
                responses of the requests that publish messages to kafka topics
            topic_offsets (Optional[dict]): offset of each topic tracked by the
                producer so far.  It is used for exact-once publishing guarantee.
        """
        self.success_topic_stats_map = {}
        requests_to_retry = self._update_success_requests_stats(
            self.requests_to_be_sent,
            responses
        )
        if self.publish_guarantee == PublishGuaranteeEnum.exact_once:
            requests_to_retry = self._verify_failed_requests(
                requests_to_retry,
                topic_offsets
            )
        self.requests_to_be_sent = requests_to_retry

    def _update_success_requests_stats(self, requests, responses):
        """Update publish stats of successful requests and return the list of
        requests that do not have success responses and need to be retried.
        """
        success_responses = {
            (r.topic, r.partition): r
            for r in responses if self._is_success_response(r)
        }

        requests_to_retry = []
        for request in requests:
            topic, partition = request.topic, request.partition

            response = success_responses.get((topic, partition))
            if not response:
                requests_to_retry.append(request)
                continue

            new_stats = _Stats(response.offset, len(request.messages))
            self._update_success_topic_stats(topic, partition, new_stats)

        return requests_to_retry

    def _is_success_response(self, response):
        """In our case, the response is either ProduceResponse (success) or
        FailedPayloadsError (failed) if no other exception is thrown.  The
        ProduceResponse should have error == 0.
        """
        return not isinstance(response, Exception) and response.error == 0

    def _update_success_topic_stats(self, topic, partition, new_stats):
        key = _TopicPartition(topic, partition)
        self.success_topic_stats_map[key] = new_stats
        self.success_topic_accum_stats_map[key] = new_stats

    def _verify_failed_requests(self, requests, topic_offsets):
        """Verify if the requests actually fail by checking the high watermark
        of the corresponding topics.  If the high watermark of a topic matches
        the number of messages in the request, the request is considered as
        successfully published, and the offset is saved in the position_data_tracker.

        If the high watermark data cannot be retrieved and it is not due to
        missing topic/partition, the request will be considered as failed but
        won't be retried because it cannot determine whether the messages are
        actually published.  Otherwise, the request will be retried.
        """
        # `get_topics_watermarks` fails all the topics if any partition leader
        # is not available, so here it checks each topic individually.
        requests_to_retry = []
        for request in requests:
            topic, partition = request.topic, request.partition
            topic_desc = "topic {} partition {} request".format(topic, partition)
            try:
                logger.debug("Verifying failed {}.".format(topic_desc))

                # try to load the metadata in case it is stale
                success = self._try_load_topic_metadata(topic)
                if not success:
                    logger.debug("Cannot load the metadata of topic {}. Skip {}."
                                 .format(topic, topic_desc))
                    continue

                published_count = self._get_published_msg_count(topic, topic_offsets)
                if len(request.messages) != published_count:
                    logger.debug(
                        "Request message count {} doesn't match actual published "
                        "message count {}. Retry {}.".format(
                            len(request.messages),
                            published_count,
                            topic_desc
                        )
                    )
                    requests_to_retry.append(request)
                    continue

                # Update stats for the request that actually succeeds
                logger.debug("{} actually succeeded.".format(topic_desc))
                new_stats = _Stats(topic_offsets[topic], published_count)
                self._update_success_topic_stats(topic, partition, new_stats)

            except LeaderNotAvailableError:
                # Topic doesn't exist yet but the broker is configured to create
                # the topic automatically. Retry the request.
                logger.debug(
                    "Topic {} doesn't exists. Retry {}.".format(topic, topic_desc)
                )
                requests_to_retry.append(request)

            except Exception:
                # Unable to get the high watermark of this topic; do not retry
                # this request since it's unclear if the messages are actually
                # successfully published.
                logger.debug(
                    "Cannot get the high watermark. Skip {}.".format(topic_desc),
                    exc_info=1
                )

        return requests_to_retry

    def _get_published_msg_count(self, topic, topic_offsets):
        published_msgs_count_map = get_actual_published_messages_count(
            self.kafka_client,
            [topic],
            topic_tracked_offset_map=topic_offsets
        )
        return published_msgs_count_map.get(topic)

    def _try_load_topic_metadata(self, topic):
        """Try to load the metadata of the topic of the given request.  It
        returns True if it succeeds, and False otherwise with one exception:
        if the topic doesn't exist but the broker is configured to create it
        automatically, it will re-raise LeaderNotAvailableError.
        """
        try:
            self.kafka_client.load_metadata_for_topics(topic)
            return True
        except LeaderNotAvailableError:
            raise
        except Exception:
            logger.exception("Failed to load metadata of topic {}.".format(topic))
            return False

    @property
    def total_published_message_count(self):
        return sum(stats.message_count
                   for stats in self.success_topic_accum_stats_map.values())

    @property
    def has_unpublished_request(self):
        """Whether any request from the initial publishing requests list hasn't
        been successfully sent."""
        request_topics_partitions = {
            (r.topic, r.partition) for r in self.initial_requests
        }
        response_topics_partitions = {
            (key.topic_name, key.partition)
            for key in self.success_topic_accum_stats_map.keys()
        }
        return not request_topics_partitions.issubset(response_topics_partitions)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline.tools.introspector.models.base import BaseIntrospectorModel


class IntrospectorNamespace(BaseIntrospectorModel):
    def __init__(self, namespace_obj, active_namespaces=None):
        super(IntrospectorNamespace, self).__init__(
            namespace_obj
        )
        self._fields = [
            'name', 'namespace_id'
        ]
        # Need to check for none in case of empty list
        if active_namespaces is not None:
            self._fields.append('active_source_count')
            self._fields.append('active_topic_count')
            self.active_source_count = 0
            self.active_topic_count = 0
            active_namespace = active_namespaces.get(self.name, None)
            if active_namespace:
                self.active_source_count = active_namespace['active_source_count']
                self.active_topic_count = active_namespace['active_topic_count']



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import cPickle
from functools import wraps


def memoized(func):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    the function is not re-evaluated.

    Based upon from http://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
    Nota bene: this decorator memoizes /all/ calls to the function.
    For a memoization decorator with limited cache size, consider:
    http://code.activestate.com/recipes/496879-memoize-decorator-function-with-cache-size-limit/
    """
    cache = {}

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        key = cPickle.dumps((args, kwargs))
        if key not in cache:
            cache[key] = func(*args, **kwargs)
        return cache[key]
    return func_wrapper



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import argparse

from data_pipeline import __version__
from data_pipeline.tools.introspector.info_command_parser import InfoCommandParser
from data_pipeline.tools.introspector.list_command_parser import ListCommandParser
from data_pipeline.tools.introspector.register_command_parser import RegisterCommandParser
from data_pipeline.tools.introspector.schema_check_command import SchemaCheckCommand


def parse_args():
    parser = argparse.ArgumentParser(
        description="data_pipeline_introspector provides ability to view the current "
        "state of the data pipeline from a top-down view of namespaces."
    )
    parser.add_argument(
        '--version',
        action='version',
        version="data_pipeline {}".format(__version__)
    )

    subparsers = parser.add_subparsers()
    ListCommandParser.add_parser(subparsers)
    InfoCommandParser.add_parser(subparsers)
    RegisterCommandParser.add_parser(subparsers)
    SchemaCheckCommand.add_parser(subparsers)
    return parser.parse_args()


def run():
    args = parse_args()
    args.command(args)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from contextlib import contextmanager

from kafka import KafkaClient
from kafka import SimpleConsumer

from data_pipeline.config import get_config
from data_pipeline.message import create_from_offset_and_message


_ONE_MEGABYTE = 1024 * 1024
logger = get_config().logger


@contextmanager
def capture_new_data_pipeline_messages(topic):
    """contextmanager that moves to the tail of the given topic, and waits to
    receive new messages, returning a function that can be called zero or more
    times which will retrieve decoded data pipeline messages from the topic.

    Returns:
        Callable[[int], List[Message]]: Function that takes a single
            optional argument, count, and returns up to count decoded data pipeline
            messages.  This function does not block, and will return however many
            messages are available immediately.  Default count is 100.
    """
    with capture_new_messages(topic) as get_kafka_messages:
        def get_data_pipeline_messages(count=100):
            kafka_messages = get_kafka_messages(count)
            return [
                create_from_offset_and_message(kafka_message)
                for kafka_message in kafka_messages
            ]

        yield get_data_pipeline_messages


@contextmanager
def capture_new_messages(topic):
    """Seeks to the tail of the topic then returns a function that can
    consume messages from that point.
    """
    with setup_capture_new_messages_consumer(topic) as consumer:
        def get_messages(count=100):
            return consumer.get_messages(count=count)

        yield get_messages


@contextmanager
def setup_capture_new_messages_consumer(topic):
    """Seeks to the tail of the topic then returns a function that can
    consume messages from that point.
    """
    kafka = KafkaClient(get_config().cluster_config.broker_list)
    group = str('data_pipeline_clientlib_test')
    consumer = SimpleConsumer(kafka, group, topic, max_buffer_size=_ONE_MEGABYTE)
    consumer.seek(0, 2)  # seek to tail, 0 is the offset, and 2 is the tail

    yield consumer

    kafka.close()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

import simplejson
from frozendict import frozendict

from data_pipeline.schematizer_clientlib.models.model_base import BaseModel
from data_pipeline.schematizer_clientlib.models.note import _Note
from data_pipeline.schematizer_clientlib.models.topic import _Topic


"""
Represent the data of an Avro schema.

Args:
    schema_id (int): The id of the avro schema.
    schema_json (dict or list): The Python representation of the avro schema.
    topic (data_pipeline.schematizer_clientlib.models.topic.Topic): The topic
        of the avro schema.
    base_schema_id (Optional[int]): The id of the base schema which this avro
        schema is changed based on.  `None` if there is no such base schema.
    status (string): The status of the schema.  It could be: "RW" (read/write),
        "R" (read-only), or "Disabled".  Read status means the schema can be
        used to deserialize messages, and Write status means the schema can be
        used to serialize messages.
    primary_keys (list): List of primary key names.
    note (Optional[data_pipeline.schematizer_clientlib.models.note.Note]): Information specified by users about the schema.
    created_at (str): The timestamp when the schema is created in ISO-8601
        format.
    updated_at (str): The timestamp when the schema is last updated in ISO-8601
        format.
"""
AvroSchema = namedtuple(
    'AvroSchema',
    ['schema_id', 'schema_json', 'topic', 'base_schema_id', 'status',
     'primary_keys', 'note', 'created_at', 'updated_at']
)


class _AvroSchema(BaseModel):
    """Internal class used to convert from/to various data structure and
    facilitate constructing the return value of schematizer functions.
    """

    def __init__(self, schema_id, schema_json, topic, base_schema_id, status,
                 primary_keys, note, created_at, updated_at):
        self.schema_id = schema_id
        self.schema_json = schema_json
        self.topic = topic
        self.base_schema_id = base_schema_id
        self.status = status
        self.primary_keys = primary_keys
        self.note = note
        self.created_at = created_at
        self.updated_at = updated_at

    @classmethod
    def from_response(cls, response):
        return cls(
            schema_id=response.schema_id,
            schema_json=simplejson.loads(response.schema),
            topic=_Topic.from_response(response.topic),
            base_schema_id=response.base_schema_id,
            status=response.status,
            primary_keys=response.primary_keys,
            note=_Note.from_response(response.note),
            created_at=response.created_at,
            updated_at=response.updated_at
        )

    def to_cache_value(self):
        return {
            'schema_id': self.schema_id,
            'schema_json': frozendict(self.schema_json),
            'topic_name': self.topic.name,
            'base_schema_id': self.base_schema_id,
            'status': self.status,
            'primary_keys': self.primary_keys,
            'note': frozendict(self.note.to_cache_value()) if self.note is not None else None,
            'created_at': self.created_at,
            'updated_at': self.updated_at
        }

    @classmethod
    def from_cache_value(cls, cache_value):
        name_only_topic = _Topic(
            topic_id=None,
            name=cache_value['topic_name'],
            source=None,
            contains_pii=None,
            cluster_type=None,
            primary_keys=None,
            created_at=None,
            updated_at=None
        )
        return cls(
            schema_id=cache_value['schema_id'],
            schema_json=cache_value['schema_json'],
            topic=name_only_topic,
            base_schema_id=cache_value['base_schema_id'],
            status=cache_value['status'],
            primary_keys=cache_value['primary_keys'],
            note=_Note.from_cache_value(cache_value['note']),
            created_at=cache_value['created_at'],
            updated_at=cache_value['updated_at']
        )

    def to_result(self):
        return AvroSchema(
            schema_id=self.schema_id,
            schema_json=self.schema_json,
            topic=self.topic.to_result(),
            base_schema_id=self.base_schema_id,
            status=self.status,
            primary_keys=self.primary_keys,
            note=self.note.to_result() if self.note is not None else None,
            created_at=self.created_at,
            updated_at=self.updated_at
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from optparse import OptionGroup

from kafka_utils.util.zookeeper import ZK
from kazoo.exceptions import NoNodeError
from yelp_batch import Batch
from yelp_batch.batch import batch_command_line_options

from data_pipeline import __version__
from data_pipeline.config import get_config
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer
from data_pipeline.servlib.config_util import load_package_config


class CompactionSetter(Batch):
    """CompactionSetter grabs all kafka topics from the schematizer, filters them by
    whether or not they have a primary_key, and then applies the 'compact' cleanup.policy setting
    on the zookeeper cluster"""

    def __init__(self):
        super(CompactionSetter, self).__init__()
        self.notify_emails = ['bam+batch@yelp.com']

    @property
    def version(self):
        """Overriding this so we'll get the clientlib version number when
        the tailer is run with --version.
        """
        return "data_pipeline {}".format(__version__)

    @batch_command_line_options
    def define_options(self, option_parser):
        opt_group = OptionGroup(option_parser, 'Compaction Setter Options')

        opt_group.add_option(
            '--dry-run',
            action="store_true",
            help="If set, will not set any topic configs, but will still run as if it is"
        )
        opt_group.add_option(
            '--config-path',
            dest='config_path',
            type='str',
            default='/nail/srv/configs/data_pipeline_tools.yaml',
            help='Config path for CompactionSetter (default: %default)'
        )
        opt_group.add_option(
            '--whitelist-topic',
            type='str',
            help='A single topic for which log compaction should be turned on'
        )
        return opt_group

    def process_commandline_options(self, args=None):
        super(CompactionSetter, self).process_commandline_options(args=args)

        load_package_config(self.options.config_path)
        self.dry_run = self.options.dry_run
        self.whitelist_topic = self.options.whitelist_topic
        self.schematizer = get_schematizer()

    def _get_all_topics_to_compact(self):
        if self.whitelist_topic:
            topics = [self.schematizer.get_topic_by_name(self.whitelist_topic)]
        else:
            topics = self.schematizer.get_topics_by_criteria()
        topics = [str(topic.name) for topic in topics]
        return self.schematizer.filter_topics_by_pkeys(topics)

    def apply_log_compaction(self, topics):
        self.log.info("Applying compaction settings on {} topics".format(len(topics)))

        compacted_topics = []
        skipped_topics = []
        missed_topics = []

        cluster = get_config().cluster_config

        with ZK(cluster) as zk:
            for topic in topics:
                try:
                    current_config = zk.get_topic_config(topic)
                    if 'cleanup.policy' not in current_config['config']:
                        # if we already have the config set or there was a
                        # manual override we don't want to set again
                        current_config['config']['cleanup.policy'] = 'compact'
                        if not self.dry_run:
                            zk.set_topic_config(topic=topic, value=current_config)
                        compacted_topics.append(topic)
                    else:
                        skipped_topics.append(topic)
                except NoNodeError:
                    missed_topics.append(topic)

        self.log_results(
            compacted_topics=compacted_topics,
            skipped_topics=skipped_topics,
            missed_topics=missed_topics
        )

    def log_results(
        self,
        compacted_topics,
        skipped_topics,
        missed_topics
    ):
        if compacted_topics:
            self.log.info(
                "Applied compaction policy to topics: {compacted_topics}".format(
                    compacted_topics=compacted_topics
                )
            )
        if skipped_topics:
            self.log.info(
                "Skipped topics (already modified cleanup policy): {skipped_topics}".format(
                    skipped_topics=skipped_topics
                )
            )
        if missed_topics:
            self.log.info(
                "Topics not found (most likely from the topics not having any recent messages):"
                "{missed_topics}".format(
                    missed_topics=missed_topics
                )
            )

    def run(self):
        self.log.info("Starting the Datapipeline Compaction Setter")

        topics = self._get_all_topics_to_compact()
        self.log.info("Found {} topics to compact".format(len(topics)))
        self.apply_log_compaction(topics)


if __name__ == '__main__':
    CompactionSetter().start()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline.tools.introspector.register.base_command import _BaseRegisterCommand


class RegisterAvroCommand(_BaseRegisterCommand):
    @classmethod
    def add_parser(cls, subparsers):
        register_avro_command_parser = subparsers.add_parser(
            "avro",
            description="Register the given avro schema to the schematizer.",
            add_help=False
        )

        cls.add_base_arguments(register_avro_command_parser)

        register_avro_command_parser.add_argument(
            "--avro-schema",
            type=str,
            required=True,
            help="The json of the avro schema."
        )

        register_avro_command_parser.add_argument(
            "--cluster-type",
            dest="cluster_type",
            default='datapipe',
            help="Kafka cluster type to connect. Defaults to datapipe. "
                 "Currently only 'datapipe' and 'scribe' cluster types are "
                 "supported."
        )

        register_avro_command_parser.add_argument(
            "--base-schema-id",
            type=int,
            default=None,
            help="The id of the original schema the new avro schema was built upon."
        )

        register_avro_command_parser.set_defaults(
            command=lambda args: cls("data_pipeline_instropsector_register_avro").run(
                args, register_avro_command_parser
            )
        )

    def process_args(self, args, parser):
        super(RegisterAvroCommand, self).process_args(args, parser)
        self.avro_schema = args.avro_schema
        self.cluster_type = args.cluster_type
        self.base_schema_id = args.base_schema_id

    def run(self, args, parser):
        self.process_args(args, parser)
        schema = self.schematizer.register_schema(
            namespace=self.namespace,
            source=self.source_name,
            schema_str=self.avro_schema,
            source_owner_email=self.source_owner_email,
            contains_pii=self.pii,
            cluster_type=self.cluster_type,
            base_schema_id=self.base_schema_id
        )
        self.print_schema(schema)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
from collections import OrderedDict


class BaseIntrospectorModel(object):
    def __init__(self, model_obj, excluded_fields=None):
        if not excluded_fields:
            excluded_fields = []
        fields_to_grab = [
            field for field in model_obj._fields if field not in excluded_fields
        ]
        for field in fields_to_grab:
            value = getattr(model_obj, field)
            if isinstance(value, datetime.datetime):
                # datetime objects are not json serializable
                value = str(value)
            setattr(self, field, value)

    def to_ordered_dict(self):
        if not hasattr(self, '_fields'):
            raise NotImplementedError(
                "Derived class does not have a defined _fields "
                "attribute to define order of fields for dict"
            )
        result_dict = OrderedDict([])
        for field in self._fields:
            result_dict[field] = getattr(self, field)
        return result_dict



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from multiprocessing import Pool

from data_pipeline._kafka_producer import _EnvelopeAndMessage
from data_pipeline._kafka_producer import _prepare
from data_pipeline._kafka_producer import LoggingKafkaProducer
from data_pipeline.config import get_config


logger = get_config().logger


class PooledKafkaProducer(LoggingKafkaProducer):
    """PooledKafkaProducer extends KafkaProducer to use a pool of subprocesses
    to schematize and pack envelopes, instead of performing those operations
    synchronously.  Parallelizing and backgrounding these expensive operations
    can result in a substantial performance improvement.

    See the Quick Start for more information about choosing an appropriate
    producer.

    TODO(DATAPIPE-171|justinc): Actually write a Quick Start
    """

    def __init__(self, *args, **kwargs):
        self.pool = Pool()
        super(PooledKafkaProducer, self).__init__(*args, **kwargs)

    def close(self):
        try:
            logger.debug("Starting to close pooled producer")
            super(PooledKafkaProducer, self).close()
            assert self.message_buffer_size == 0
            logger.debug("Closing the pool")
            self.pool.close()
            logger.debug("Pool is closed.")
        except:
            logger.error("Exception occurred when closing pooled producer.")
            raise
        finally:
            # The processes in the pool should be cleaned up in all cases. The
            # exception will be re-thrown if there is one.
            #
            # Joining pools can be flaky in CPython 2.6, and the message buffer
            # size is zero here, so terminating the pool is safe and ensure that
            # join always works.
            self.pool.terminate()
            self.pool.join()

    def _prepare_message(self, message):
        """This happens in the pool, so this is a noop"""
        return message

    def _generate_prepared_topic_and_messages(self):
        # The setup here isn't great, it's probably worth switching this to
        # keep the buffer in an array, then map it here.  It'd also be worth
        # looking at pipelining this, so there would be a regular buffer, and a
        # buffer that was being prepared.
        #
        # That would look like:
        #
        # publish -> accumulate in buffer -> move to prep area -> async prepare
        #   -> publish to kafka
        #
        # clin proposed an alternative approach:
        # I personally would prefer to pipeline this, and spawn certain number
        # of workers doing schematizing and packing messages (listen to regular
        # queue and move items from it to schematized queue), and another
        # certain number of workers sending messages to kafka (listen to
        # schematized queue and move items from it to kafka topic). The number
        # of workers can be configurable so we can treak performance if needed.
        # One benefit is we can build some metric from it and see the throughput
        # or identify where the bottleneck is if there is one.
        #
        # Also, in this way, it seems the schematizing workers don't have to
        # work in bulk and can continuously schematize and pack the incoming raw
        # messages, and move them into schemaized/packed queue (if there are
        # free workers). The send-requests workers can then send the messages
        # in bulk or every certain amount of time. The down side is this is a
        # more complicated approach.
        topics_and_messages_result = [
            (topic, self.pool.map_async(
                _prepare,
                [
                    _EnvelopeAndMessage(envelope=self.envelope, message=message)
                    for message in messages
                ]
            )) for topic, messages in self.message_buffer.iteritems()
        ]

        return [(topic, messages_result.get()) for topic, messages_result in topics_and_messages_result]
